import textwrap

from ..definition import (
    Dimension,
    Measure,
    Model,
    Registry,
    Relationship,
    RelationshipType,
)


def register_metrics(registry: Registry, catalog_name: str = "iceberg"):
    # ============================================"
    # METRICS MODELS
    # ============================================"

    registry.register(
        Model(
            name="metrics",
            table=f"{catalog_name}.oso.metrics_v0",
            description=textwrap.dedent(
                """
                Each row represents metadata about a unique metric. This data can be
                used to analyze trends and performance indicators across various
                open source projects, supporting business decisions around developer
                engagement, funding, and project health. This table is necessary for
                doing keyword searches to discover relevant metrics ids, and joining
                on the other metrics tables (eg, key_metrics_by_project and
                timeseries_metrics_by_project). Business questions that can be
                answered include: Which daily metrics are available for "BASE"? What
                are the available metric ids for GITHUB_stars? Which metrics are
                relevant to contract artifacts on "OPTIMISM"?'
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="metric_source",
                    description="The source of the metric, typically OSO",
                    column_name="metric_source",
                ),
                Dimension(
                    name="metric_namespace",
                    description="The namespace grouping of the metric",
                    column_name="metric_namespace",
                ),
                Dimension(
                    name="metric_name",
                    description="The name of the metric",
                    column_name="metric_name",
                ),
                Dimension(
                    name="display_name",
                    description="A human-readable name for the metric",
                    column_name="display_name",
                ),
                Dimension(
                    name="description",
                    description="A description of the metric",
                    column_name="description",
                ),
                Dimension(
                    name="rendered_sql",
                    description="The SQL query used to compute the metric",
                    column_name="rendered_sql",
                ),
                Dimension(
                    name="sql_source_path",
                    description="The path to the SQL file that defines the metric",
                    column_name="sql_source_path",
                ),
                Dimension(
                    name="aggregation_function",
                    description="The aggregation function used for the metric",
                    column_name="aggregation_function",
                ),
            ],
            primary_key="metric_id",
            measures=[
                Measure(
                    name="count",
                    description="The number of metrics",
                    query="COUNT(self.metric_id)",
                ),
                Measure(
                    name="distinct_count",
                    description="The number of distinct metrics",
                    query="COUNT(DISTINCT self.metric_id)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="timeseries_metrics_by_artifact",
            table=f"{catalog_name}.oso.timeseries_metrics_by_artifact_v0",
            description=textwrap.dedent(
                """
                Many-to-many table between metrics, artifacts, and dates. Each row
                represents the value of a specific metric for a given artifact on a
                particular date, capturing how that metric changes over time. This
                is one of the most important models for OSO and should be used
                frequently in analysis. However, in order to use this table
                correctly, it is necessary to join on metrics and artifacts
                (or other artifact-related tables). Business questions that can be
                answered include: Which artifacts have the most active developers?
                How many forks does this artifact have?
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="artifact_id",
                    description="The unique identifier of the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="sample_date",
                    description="The date when the metric was sampled",
                    column_name="sample_date",
                ),
                Dimension(
                    name="amount",
                    description="The value of the metric",
                    column_name="amount",
                ),
                Dimension(
                    name="unit",
                    description="The unit of the metric",
                    column_name="unit",
                ),
            ],
            primary_key=["metric_id", "artifact_id", "sample_date"],
            measures=[
                Measure(
                    name="sum",
                    description="The total sum of artifact metrics",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg",
                    description="The average of artifact metrics",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="count",
                    description="The count of metric records",
                    query="COUNT(*)",
                ),
            ],
            relationships=[
                Relationship(
                    source_foreign_key="artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="metric_id",
                    ref_model="metrics",
                    ref_key="metric_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="timeseries_metrics_by_project",
            table=f"{catalog_name}.oso.timeseries_metrics_by_project_v0",
            description=textwrap.dedent(
                """
                Many-to-many table between metrics, projects, and dates. Each row
                represents the value of a specific metric for a given project on a
                particular date, capturing how that metric changes over time. This
                is one of the most important models for OSO and should be used
                frequently in analysis. However, in order to use this table
                correctly, it is necessary to join on metrics and projects
                (or other project-related tables). Business questions that can be
                answered include: Which projects have the most active developers?
                How many forks does this project have?
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="project_id",
                    description="The unique identifier of the project",
                    column_name="project_id",
                ),
                Dimension(
                    name="sample_date",
                    description="The date when the metric was sampled",
                    column_name="sample_date",
                ),
                Dimension(
                    name="amount",
                    description="The value of the metric",
                    column_name="amount",
                ),
                Dimension(
                    name="unit",
                    description="The unit of the metric",
                    column_name="unit",
                ),
            ],
            primary_key=["metric_id", "project_id", "sample_date"],
            measures=[
                Measure(
                    name="sum",
                    description="The total sum of project metrics",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg",
                    description="The average of project metrics",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="count",
                    description="The count of metric records",
                    query="COUNT(*)",
                ),
            ],
            relationships=[
                Relationship(
                    source_foreign_key="project_id",
                    ref_model="projects",
                    ref_key="project_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="metric_id",
                    ref_model="metrics",
                    ref_key="metric_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="timeseries_metrics_by_collection",
            table=f"{catalog_name}.oso.timeseries_metrics_by_collection_v0",
            description=textwrap.dedent(
                """
                Many-to-many table between metrics, collections, and dates. Each row
                represents the value of a specific metric for a given collection on a
                particular date, capturing how that metric changes over time. This
                is one of the most important models for OSO and should be used
                frequently in analysis. However, in order to use this table
                correctly, it is necessary to join on metrics and collections
                (or other collection-related tables). Business questions that can be
                answered include: Which collections have the most active developers?
                How many forks does this collection have?
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="collection_id",
                    description="The unique identifier of the collection",
                    column_name="collection_id",
                ),
                Dimension(
                    name="sample_date",
                    description="The date when the metric was sampled",
                    column_name="sample_date",
                ),
                Dimension(
                    name="amount",
                    description="The value of the metric",
                    column_name="amount",
                ),
                Dimension(
                    name="unit",
                    description="The unit of the metric",
                    column_name="unit",
                ),
            ],
            primary_key=["metric_id", "collection_id", "sample_date"],
            measures=[
                Measure(
                    name="sum",
                    description="The total sum of collection metrics",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg",
                    description="The average of collection metrics",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="count",
                    description="The count of metric records",
                    query="COUNT(*)",
                ),
            ],
            relationships=[
                Relationship(
                    source_foreign_key="collection_id",
                    ref_model="collections",
                    ref_key="collection_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="metric_id",
                    ref_model="metrics",
                    ref_key="metric_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="key_metrics_by_artifact",
            table=f"{catalog_name}.oso.key_metrics_by_artifact_v0",
            description=textwrap.dedent(
                """
                Key metrics for artifacts. Each row represents the latest value of 
                a specific metric for a given artifact.
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="artifact_id",
                    description="The unique identifier of the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="sample_date",
                    description="The date when the metric was sampled",
                    column_name="sample_date",
                ),
                Dimension(
                    name="amount",
                    description="The value of the metric for the artifact",
                    column_name="amount",
                ),
                Dimension(
                    name="unit",
                    description="The unit of measurement for the metric",
                    column_name="unit",
                ),
            ],
            primary_key=["metric_id", "artifact_id"],
            measures=[
                Measure(
                    name="sum",
                    description="The total sum of artifact metrics",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg",
                    description="The average of artifact metrics",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="count",
                    description="The count of metric records",
                    query="COUNT(*)",
                ),
            ],
            relationships=[
                Relationship(
                    source_foreign_key="artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="metric_id",
                    ref_model="metrics",
                    ref_key="metric_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="key_metrics_by_project",
            table=f"{catalog_name}.oso.key_metrics_by_project_v0",
            description=textwrap.dedent(
                """
                Key metrics for projects. Each row represents the latest value of 
                a specific metric for a given project.
                """
            ),
            dimensions=[
                Dimension(
                    name="metric_id",
                    description="The unique identifier of the metric",
                    column_name="metric_id",
                ),
                Dimension(
                    name="project_id",
                    description="The unique identifier of the project",
                    column_name="project_id",
                ),
                Dimension(
                    name="sample_date",
                    description="The date when the metric was sampled",
                    column_name="sample_date",
                ),
                Dimension(
                    name="amount",
                    description="The value of the metric for the project",
                    column_name="amount",
                ),
                Dimension(
                    name="unit",
                    description="The unit of measurement for the metric",
                    column_name="unit",
                ),
            ],
            primary_key=["metric_id", "project_id"],
            measures=[
                Measure(
                    name="sum",
                    description="The total sum of project metrics",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg",
                    description="The average of project metrics",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="count",
                    description="The count of metric records",
                    query="COUNT(*)",
                ),
            ],
            relationships=[
                Relationship(
                    source_foreign_key="project_id",
                    ref_model="projects",
                    ref_key="project_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="metric_id",
                    ref_model="metrics",
                    ref_key="metric_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )



import typing as t

if t.TYPE_CHECKING:
    from .definition import AttributePath

class InvalidAttributeReferenceError(Exception):
    """Exception raised when an invalid attribute reference is encountered."""

    def __init__(self, reference: "AttributePath", message: str = "intermediate attributes must be relationship"):
        super().__init__(f"Invalid attribute reference `{reference}`: {message}")
        self.reference = reference


"""Semantic Layer Type Definitions

The semantic layer provides a structured way to access any data model. This is
heavily inspired by prior work with cube.dev, looker, and some new work we've
seen in Snowflake. This semantic layer library is designed to generate SQL
queries as opposed to directly running them. It is powered by sqlglot and
therefore the final sql can render to any dialect supported by sqlglot.

See our documentation for more information at https://docs.opensource.observer.
"""

from __future__ import annotations

import hashlib
import logging
import textwrap
import typing as t
from collections import deque
from enum import Enum
from graphlib import TopologicalSorter

from pydantic import BaseModel, ConfigDict, Field, model_validator
from sqlglot import exp, parse_one

from .errors import InvalidAttributeReferenceError
from .utils import exp_to_str, hash_expressions

logger = logging.getLogger(__name__)


def coerce_to_tables_column(ref: str, expected_table: str = "self") -> exp.Expression:
    """Coerces a string reference to a column in a specific table."""
    column = exp.to_column(ref)

    if column.table == expected_table:
        return column
    if column.table == "":
        return exp.to_column(f"{expected_table}.{ref}")

    raise ValueError(f"{ref} is not a valid {expected_table} reference")


class QueryRegistry(t.Protocol):
    def __init__(self, registry: Registry): ...
    def select(self, *selects: str) -> t.Self: ...
    def where(self, *filters: str) -> t.Self: ...
    def limit(self, limit: int) -> t.Self: ...
    def cte(self, name: str, query: t.Self) -> t.Self: ...
    def to_model(self, name: str) -> Model: ...
    def build(self) -> exp.Expression: ...


class RegistryGraph:
    """A graph of models and views that can be used to generate SQL queries.

    This is used to determine the order in which models and views should be joined.
    """

    adjacency_map: t.Dict[str, t.Set[str]]

    def __init__(self):
        self.adjacency_map = {}

    def add(self, model: "Model"):
        """Inserts the model into the registry and updates the adjacency map."""

        # Register the model under the given name
        self.adjacency_map[model.name] = set()

        # Gets all the names of related models
        reference_names = list(map(lambda x: x.ref_model, model.relationships))

        for reference in reference_names:
            self.adjacency_map.setdefault(reference, set())
            self.adjacency_map[reference].add(model.name)
            self.adjacency_map[model.name].add(reference)

    def find_best_join_tree(self, references: list[AttributePath]):
        """Finds the join tree with the least amount of nodes (models) for the given references."""
        join_trees = sorted(
            [
                tree
                for tree in (
                    self.build_join_tree(root.base_model, references)
                    for root in references
                )
                if tree is not None
            ],
            key=lambda x: len(x),
        )
        if len(join_trees) == 0:
            raise ModelHasNoJoinPath(
                "No join tree found for the given references. Ensure that the references are valid and that the models are registered."
            )
        return join_trees[0]

    def build_join_tree(self, root: str, references: list[AttributePath]):
        # Build a tree rooted at root
        queue = deque([root])
        parents = {root: root}

        while queue:
            current = queue.popleft()
            for child in self.adjacency_map.get(current, set()):
                if child not in parents:
                    parents[child] = current
                    queue.append(child)

        # Only include in the tree models/paths that are referenced
        filtered_parents = {root: root}
        for reference in references:
            traverser = reference.traverser()
            while True:
                curr_model = traverser.current_model_name
                while curr_model != root and curr_model not in filtered_parents:
                    if curr_model not in parents:
                        return None
                    filtered_parents[curr_model] = parents[curr_model]
                    curr_model = parents[curr_model]
                if not traverser.next():
                    break

        return JoinTree(root, filtered_parents)


class JoinTree:
    root: str
    parents: dict[str, str]

    def __init__(self, root: str, parents: dict[str, str]):
        self.root = root
        self.parents = parents

        self.depths = {}
        self._calculate_depths()

    def __len__(self):
        return len(self.parents)

    def _calculate_depths(self):
        def _get_depth(node: str) -> int:
            if node in self.depths:
                return self.depths[node]
            if node == self.root:
                self.depths[node] = 0
                return 0

            parent = self.parents[node]
            depth = _get_depth(parent) + 1
            self.depths[node] = depth
            return depth

        for node in self.parents:
            _get_depth(node)

    def get_path(self, from_node: str, to_node: str):
        if from_node not in self.parents:
            raise ValueError(f"Node {from_node} not found in join tree")
        if to_node not in self.parents:
            raise ValueError(f"Node {to_node} not found in join tree")
        if from_node == to_node:
            return []

        from_depth = self.depths[from_node]
        to_depth = self.depths[to_node]
        current_from = from_node
        current_to = to_node

        # Bring both nodes to the same depth level
        # Move the deeper node up until both are at the same depth
        from_path = []
        to_path = []

        while from_depth > to_depth:
            from_path.append(current_from)
            current_from = self.parents[current_from]
            from_depth -= 1

        while to_depth > from_depth:
            to_path.append(current_to)
            current_to = self.parents[current_to]
            to_depth -= 1

        # Now both nodes are at the same depth, move both up until they meet (LCA)
        while current_from != current_to:
            from_path.append(current_from)
            to_path.append(current_to)
            current_from = self.parents[current_from]
            current_to = self.parents[current_to]

        lca = current_from

        to_path.reverse()
        result = from_path + [lca] + to_path

        return result


class BoundModelAttribute(t.Protocol):
    def to_query_component(
        self,
        traverser: "AttributePathTraverser",
        original: "AttributePath | str",
        registry: "Registry",
    ) -> "QueryComponent":
        """Returns the QueryComponent for the bound query part"""
        ...

    def validate(self): ...


class UnboundModelAttribute(t.Protocol):
    def bind(
        self, model: "Model", dependencies: dict[str, "AttributePath"]
    ) -> "BoundModelAttribute":
        """Binds the query part to the given model and returns a attribute"""
        ...

    def attribute_references(self) -> t.List["AttributePath"]:
        """Returns the references for model attributes"""
        ...


Q = t.TypeVar("Q", bound=QueryRegistry)


class Registry(t.Generic[Q]):
    """A registry for semantic models and views. The generic type Q must follow
    the QueryRegistry protocol. By default this relies on the QueryBuilder class
    available in this library.

    Args:
        query_builder: The QueryRegistry compliant class constructor to use for
        the registry that must follow the QueryRegistry protocol. If not
        provided, defaults to the QueryBuilder class.
    """

    models: t.Dict[str, "Model"]
    dag: RegistryGraph
    views: t.Dict[str, "View"]

    def __init__(self, query_builder: type[Q] | None = None):
        from .query import QueryBuilder

        self.query_builder: type[Q] = query_builder or QueryBuilder  # type: ignore[assignment]
        self.models = {}
        self.views = {}
        self.sorted_model_names = None
        self.dag = RegistryGraph()

    def register(self, model: "Model", repoen: bool = False):
        """Inserts the model into the registry and updates the adjacency map.

        Cycles are not allowed in the registry. If a cycle is detected, an error
        will be raised on later calls
        """

        self.models[model.name] = model
        # Register the model under the given name
        self.dag.add(model)

    def register_view(self, view: "View"):
        self.views[view.name] = view

    def get_model(self, name: str) -> "Model":
        return self.models[name]

    def get_view(self, name: str) -> "View":
        return self.views[name]

    def query_from_object(self, semantic_query: "SemanticQuery"):
        """Allows initializing a query builder from a SemanticQuery object."""

        query = self.select(*semantic_query.selects)
        if semantic_query.filters:
            query = query.where(*semantic_query.filters)
        if semantic_query.limit:
            query = query.limit(semantic_query.limit)
        return query

    def cte(self, name: str, query: Q):
        """Returns a new query builder for the registry with a subquery"""

        query_builder = self.query_builder(self).cte(name, query)
        return query_builder

    def select(self, *selects: str):
        """Returns a new query builder for the registry"""

        query_builder = self.query_builder(self)
        for select in selects:
            if not isinstance(select, str):
                raise ValueError(f"Select must be a string, got {type(select)}")
        query_builder = query_builder.select(*selects)
        return query_builder

    def describe(self) -> str:
        """Returns a description of the registry used for an LLM prompt"""

        description = "This is the semantic model available to query\n"

        description += "# All Available Models\n"
        for model in self.models.values():
            description += model.describe() + "\n"

        return description

    def expand_reference(
        self,
        reference: "AttributePath",
    ) -> t.List["AttributePath"]:
        """Given a reference, walk it and resolve all of the related references"""
        final = reference.final_attribute()

        model = self.get_model(exp_to_str(final.table))
        sub_references = model.get_attribute_references(exp_to_str(final.this))
        if len(sub_references) == 0:
            return [reference]
        else:
            traverser = reference.traverser()
            traverser.to_end()

            expanded_references: t.List[AttributePath] = [reference]

            for sub_reference in sub_references:
                # Rewrite the sub reference to the current model
                expanded_references.extend(
                    [
                        traverser.create_scoped_reference(ref)
                        for ref in self.expand_reference(sub_reference)
                    ]
                )
            return expanded_references

    def get_attribute_from_reference(
        self, reference: "AttributePath"
    ) -> "BoundDimension | BoundMeasure | BoundRelationship":
        attr = reference.final_attribute()

        model = self.get_model(exp_to_str(attr.table))
        attribute_name = exp_to_str(attr.this)

        return model.get_attribute(attribute_name)


class GenericExpression(BaseModel):
    """Specifically not the sqlglot expression type, but a generic internal type
    for expressions

    When defining a model the queries that make up a model all inherit from this
    base class. Additionally, all expressions within a query (selects and
    filters) also inherit from this class.

    There are two main types of expressions:

    1. RawExpression: This is a raw SQL expression. Any references to columns is
       considered should specifically be references to the _actual_ column names
       on the underlying table.
    2. SemanticExpression: This is a semantic expression that is in the semantic
       context and can reference models and attributes. When used in a dimension
       or measure it may only reference `self`. In queries, it may reference any
       model attribute available in the registry.

    In both cases, the query string is processed at the time of instantiation
    and all references to columns or other semantic layer attributes are
    converted to sqlglot exp.Anonymous (which represent some unknown sql
    function if rendered to sql). This is done to allow for us to preprocess all
    of the available references in an expression and replace them with the
    correct table.column reference in the final query that is generated. For
    RawExpression, the anonymous function is `$column_ref`. For
    SemanticExpression, the anonymous function is `$semantic_ref`.

    To get the final SQL expression for a given GenericExpression, you must call
    the `resolve` method. This will convert all of the exp.Anonymous expressions
    that match the `$semantic_ref` and `$column_ref` into the correct
    table.column references for a given sql query.
    """

    query: str

    @model_validator(mode="after")
    def validate_query(self):
        """Validates that the query is a valid SQL expression"""
        try:
            self._expression = self.to_sqlglot_expression()
        except Exception as e:
            raise ValueError(f"Invalid SQL expression: {e}")

        return self

    @property
    def expression(self) -> exp.Expression:
        return self._expression

    def to_sqlglot_expression(self) -> exp.Expression:
        """Returns the SQL expression for the generic expression"""
        raise NotImplementedError(
            "This method should be implemented by subclasses to return the SQL expression for the generic expression"
        )

    def columns(self) -> t.List[exp.Column]:
        """Returns the columns in the expression. By default a generic
        expressions are _not_ allowed to reference columns"""
        raise ValueError(
            "GenericExpression columns method does not allow self references by default. Use allow_self=True to allow self references."
        )

    def references(self, allow_self: bool = False) -> t.List["AttributePath"]:
        """Returns the semantic references in the expression

        Args:
            allow_self: If True, allows references to the `self` model. Defaults to False.

        Returns:
            A list of AttributePath objects representing the semantic references in the expression.
        """

        raise NotImplementedError(
            "This method should be implemented by subclasses to return the semantic references in the expression"
        )

    def placeholder_expression_to_attribute_path(
        self, node: exp.Anonymous
    ) -> "AttributePath":
        """Converts a placeholder expression to an AttributePath.

        This is used to convert the $semantic_ref and $column_ref expressions
        into AttributePath objects.
        """
        if not node.expressions:
            raise ValueError("Placeholder expression must have at least one expression")
        reference_literal = exp_to_str(node.expressions[0])
        try:
            return AttributePath.from_string(reference_literal)
        except InvalidAttributeReferenceError as e:
            raise ValueError(f"Invalid semantic reference: {e}")

    def resolve(
        self,
        registry: Registry,
        self_model_name: str = "",
        traverser: AttributePathTraverser | None = None,
    ) -> exp.Expression:
        """Returns the SQL expression for an expression. This generic resolve
        method turns $semantic_ref, and $column_ref anonymous expressions into
        executable sql based on the current registry and traverser.

        Args:
            registry: The registry
            self_model_name: The name of the model that this expression is being
                resolved for. This is used to determine the current model in the
                traverser and should generally only be set by a BoundDimension,
                BoundMeasure, or BoundRelationship.
            traverser: The traverser to use for resolving the expression. If not
                provided, a new traverser will be created from the root.

        Returns:
            The resolved SQL expression as a sqlglot Expression object.
        """
        if not traverser:
            traverser = AttributePathTraverser.from_root()

        def convert_column(node: exp.Expression):
            if isinstance(node, exp.Anonymous):
                anonymous_function_name = exp_to_str(node.this)
                if anonymous_function_name not in ["$semantic_ref", "$column_ref"]:
                    # If this is not a semantic reference or column reference, we can skip it
                    return node

                # Get the model reference from the anonymous expression
                reference = self.placeholder_expression_to_attribute_path(node)
                reference = traverser.create_scoped_reference(reference)

                column_traverser = reference.traverser()
                column_traverser.to_end()
                current_model_name = column_traverser.current_model_name
                if current_model_name == "self":
                    # If the current model is self, we need to use the self_model_name
                    current_model_name = self_model_name
                model = registry.get_model(current_model_name)

                if anonymous_function_name == "$semantic_ref":
                    # If this is a semantic reference then we need to recursively resolve it
                    attribute = model.get_attribute(
                        column_traverser.current_attribute_name
                    )
                    match attribute:
                        case BoundDimension():
                            resolved = attribute.resolve(
                                registry=registry,
                                traverser=column_traverser,
                            )
                            return resolved
                        case BoundMeasure():
                            resolved = attribute.resolve(
                                registry=registry,
                                traverser=column_traverser,
                            )
                            return resolved
                        case BoundRelationship():
                            resolved = attribute.resolve(
                                registry=registry,
                                traverser=column_traverser,
                            )
                            return resolved

                elif anonymous_function_name == "$column_ref":
                    # If this is a column reference then we need to simply resolve to the model's column
                    col = exp.to_column(
                        f"{column_traverser.alias(current_model_name)}.{column_traverser.current_attribute_name}",
                    )
                    return col
            return node

        return self._expression.transform(convert_column)

    @classmethod
    def is_ref_node(cls, node: exp.Expression) -> bool:
        """Checks if the node is an anonymous reference node.

        This is used to determine if the node is a semantic reference or a column
        reference.
        """
        return isinstance(node, exp.Anonymous) and (
            exp_to_str(node.this) == "$semantic_ref"
            or exp_to_str(node.this) == "$column_ref"
        )

    @classmethod
    def is_column_ref_node(cls, node: exp.Expression) -> bool:
        """Checks if the node is a column reference node.

        This is used to determine if the node is a column reference.
        """
        return (
            isinstance(node, exp.Anonymous) and exp_to_str(node.this) == "$column_ref"
        )

    @classmethod
    def is_semantic_ref_node(cls, node: exp.Expression) -> bool:
        """Checks if the node is a semantic reference node.

        This is used to determine if the node is a semantic reference.
        """
        return (
            isinstance(node, exp.Anonymous) and exp_to_str(node.this) == "$semantic_ref"
        )


class SemanticExpression(GenericExpression):
    """A string expression that is in the semantic context"""

    def to_sqlglot_expression(self) -> exp.Expression:
        """Transforms the query into a sqlglot expression"""

        def transform(node: exp.Expression) -> exp.Expression:
            if is_expression_attribute_reference(node):
                # For anything that's an attribut path we need to resolve it
                if isinstance(node, exp.Column):
                    # if it's a column this is a simple resolution simply to the column via the registry
                    logger.debug(f"Found column reference: {node}")
                    model_name = exp_to_str(node.table)
                    column_name = exp_to_str(node.this)
                    reference = AttributePath.from_string(f"{model_name}.{column_name}")
                else:
                    logger.debug(f"Found non column reference: {node}")
                    # Check for self in reference path
                    reference_path = node.sql(dialect="duckdb").split("->")
                    for i in range(len(reference_path)):
                        ref = reference_path[i]
                        as_column = exp.to_column(ref)
                        model_name = exp_to_str(as_column.table)

                        if model_name == "self":
                            reference_path[i] = (
                                f"{model_name}.{exp_to_str(as_column.this)}"
                            )
                    reference = AttributePath(path=reference_path)

                refs_as_literals = exp.Literal.string(str(reference))

                return exp.Anonymous(
                    this="$semantic_ref", expressions=[refs_as_literals]
                )
            return node

        return parse_one(self.query).transform(transform)

    def references(self, allow_self: bool = False) -> t.List["AttributePath"]:
        """Returns the semantic references in the expression"""
        glot_expression = self._expression
        all_reference_placeholders = glot_expression.find_all(exp.Anonymous)
        references: t.List[AttributePath] = []
        for reference_placeholder in all_reference_placeholders:
            if reference_placeholder.this == "$semantic_ref":
                # This is a semantic reference, we need to parse it
                if not reference_placeholder.expressions:
                    raise ValueError(
                        "Semantic reference must have at least one expression"
                    )
                reference_literal = exp_to_str(reference_placeholder.expressions[0])
                try:
                    reference = AttributePath.from_string(reference_literal)
                except InvalidAttributeReferenceError as e:
                    raise ValueError(f"Invalid semantic reference: {e}")
                if not allow_self and reference.is_self_reference():
                    raise ValueError(
                        f"SemanticExpression {self.query} not allowed to reference `self` model. Only valid in Dimension, Measure, and Relationship definitions."
                    )
                references.append(reference)
        return references


class RawExpression(GenericExpression):
    """A raw SQL expression that is not in the semantic context.

    This is used to pass raw SQL expressions to the semantic layer without
    parsing them. However, the special table `self` is used to refer to the
    current model's table.
    """

    query: str

    @model_validator(mode="after")
    def validate_only_self_references(self):
        """Validates that the expression only references `self`

        RawExpressions have no access to the semantic layer and therefore
        cannot reference other models. They can only reference the current model's
        table.
        """
        for column in self.columns():
            if exp_to_str(column.table) != "self":
                raise ValueError(
                    f"Raw expression {self.query} cannot reference another model. Found reference to {column.table}. Only use `self`"
                )
        return self

    def to_sqlglot_expression(self) -> exp.Expression:
        def convert_column(node: exp.Expression):
            if isinstance(node, exp.Column):
                return exp.Anonymous(
                    this="$column_ref",
                    expressions=[exp.Literal.string(f"{node.table}.{node.this}")],
                )
            return node

        return parse_one(self.query).transform(convert_column)

    def columns(self) -> t.List[exp.Column]:
        """Returns the columns in the expression"""
        expression = parse_one(self.query)
        return list(expression.find_all(exp.Column))

    def references(self, allow_self: bool = False) -> t.List["AttributePath"]:
        return []


def raw_exp(query: str) -> RawExpression:
    """Creates a raw expression from a string"""
    return RawExpression(query=query)


def semantic_exp(query: str) -> SemanticExpression:
    """Creates a semantic expression from a string"""
    return SemanticExpression(query=query)


class MeasureOperationType(str, Enum):
    """The type of operation to perform on the measure.

    This is used to determine how to aggregate the measure.
    """

    SUM = "sum"
    AVG = "avg"
    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"
    MAX = "max"
    MIN = "min"
    FIRST = "first"
    LAST = "last"
    PERCENTILE = "percentile"
    PERCENTILE_CONT = "percentile_cont"
    PERCENTILE_DISC = "percentile_disc"


class Measure(BaseModel):
    name: str
    description: str = ""
    query: RawExpression | SemanticExpression | str = ""

    @model_validator(mode="after")
    def process_metric(self):
        if isinstance(self.query, str):
            expression = SemanticExpression(query=self.query)
        else:
            expression = self.query

        self._expression = expression

        return self

    @property
    def columns(self) -> t.List[exp.Column]:
        """Returns the columns in the expression"""
        if isinstance(self._expression, RawExpression):
            return list(self._expression.columns())
        else:
            return []

    def bind(self, model: "Model") -> "BoundMeasure":
        """Returns the bound metric for the given model"""

        return BoundMeasure(
            model=model,
            measure=self,
        )

    def resolve(
        self,
        self_model_name: str,
        registry: Registry,
        traverser: "AttributePathTraverser | None" = None,
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the measure."""

        return self._expression.resolve(
            traverser=traverser,
            self_model_name=self_model_name,
            registry=registry,
        )

    def references(self, allow_self: bool = False) -> t.List["AttributePath"]:
        """Returns the semantic references in the measure expression"""
        return self._expression.references(allow_self=allow_self)


class BoundMeasure:
    model: "Model"
    measure: Measure

    def __init__(self, model: "Model", measure: Measure):
        self.model = model
        self.measure = measure

    def resolve(
        self,
        registry: Registry,
        traverser: "AttributePathTraverser | None" = None,
    ) -> exp.Expression:
        """Returns the QueryComponent for the measure"""

        return self.measure.resolve(
            traverser=traverser,
            self_model_name=self.model.name,
            registry=registry,
        )

    def references(self) -> t.List["AttributePath"]:
        """Returns the semantic references in the dimension expression"""
        return [
            ref.rewrite_self(self.model.name)
            for ref in self.measure.references(allow_self=True)
        ]


class Select(BaseModel):
    """Represents the select part of a query. This should just be a semantic reference"""

    query: SemanticExpression | str = Field(
        description="A semantic expression for the select part of the query"
    )

    @model_validator(mode="after")
    def process_select(self):
        if isinstance(self.query, str):
            expression = SemanticExpression(query=self.query)
        else:
            expression = self.query

        references = expression.references(allow_self=False)

        # We currently don't support adding constants as a select expression
        if len(references) != 1:
            raise ValueError(
                f"Select expression must be a single semantic reference, got {len(references)} references: {references}"
            )

        # Ensure that the sqlglot expression is only a single reference or a single
        # reference and an alias
        sqlglot_expr = expression.to_sqlglot_expression()
        if not GenericExpression.is_semantic_ref_node(sqlglot_expr):
            if not isinstance(sqlglot_expr, exp.Alias):
                raise ValueError(
                    f"Select expression must be a semantic reference or a semantic reference with an alias, got {type(sqlglot_expr)}"
                )

        self._expression = expression

        return self

    @property
    def expression(self) -> SemanticExpression:
        """Returns the semantic expression for the select part of the query"""
        return self._expression

    def resolve(
        self, registry: Registry, traverser: AttributePathTraverser | None = None
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the select part of the query."""

        return self._expression.resolve(
            traverser=traverser,
            self_model_name="",
            registry=registry,
        )

    def references(self) -> t.List["AttributePath"]:
        """Returns the semantic references in the select expression"""
        return self._expression.references(allow_self=False)

    def is_aggregate(self, registry: Registry) -> bool:
        """Returns whether the select expression is an aggregate function."""
        # If this references a measure, then it is an aggregate function
        references = self.references()
        for reference in references:
            attribute = registry.get_attribute_from_reference(reference)
            if isinstance(attribute, BoundMeasure):
                return True
        return False

    def alias(self):
        """Returns the alias for the select expression"""
        # The alias is the last part of the reference path
        # Check the expression if it explicitly has an alias
        sqlglot_expr = self._expression.to_sqlglot_expression()
        if isinstance(sqlglot_expr, exp.Alias):
            # If the expression is an alias, we use that as the alias

            return exp_to_str(sqlglot_expr.alias)
        else:
            reference = self.references()[0]
            return reference.to_select_alias()


class Filter(BaseModel):
    """Represents the filter part of a query.

    Filters' queries are only SemanticExpressions.
    """

    query: SemanticExpression | str = Field(
        description="A semantic expression for the filter"
    )

    @model_validator(mode="after")
    def process_filter(self):
        if isinstance(self.query, str):
            expression = SemanticExpression(query=self.query)
        else:
            expression = self.query

        self._expression = expression

        return self

    @property
    def expression(self) -> SemanticExpression:
        """Returns the semantic expression for the filter"""
        return self._expression

    def resolve(
        self,
        registry: Registry,
        traverser: AttributePathTraverser | None = None,
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the filter."""

        return self._expression.resolve(
            traverser=traverser,
            self_model_name="",
            registry=registry,
        )

    def references(self) -> t.List["AttributePath"]:
        """Returns the semantic references in the dimension expression"""
        return self._expression.references(allow_self=False)

    def is_aggregate(self, registry: Registry) -> bool:
        """Returns whether the filter expression is an aggregate function."""
        # If this references a measure, then it is an aggregate function (this likely needs to be fixed but)
        references = self.references()
        for reference in references:
            attribute = registry.get_attribute_from_reference(reference)
            if isinstance(attribute, BoundMeasure):
                return True
        return False

    def alias(self) -> str:
        return ""


class Dimension(BaseModel):
    """A dimension is a column or expression that must resolve to a column in
    the table.

    The dimension can either be a simple column reference or a non-aggregating
    expression.
    """

    name: str
    description: str = ""
    query: RawExpression | SemanticExpression | str = Field(
        default="",
        description=textwrap.dedent(
            """
        A SQL expression for the dimension. If using a "RawExpression", it any
        columns in the query are listed as expected columns in the table. If
        using a SemanticExpression, it must resolve to a different dimension in
        the model. It may _not_ resolve to a measure or relationship. If this is
        not provided, the dimension will default to the column with the same
        name as the dimension.

        Examples:
        
            Let's assume we initialize a model with an attribute `time` that has
            the raw expression `RawExpression(query='self.time')`. Let's also
            assume this is stored as a unix timestamp. If we wanted to have
            another attribute `date` we could reference the time column OR the
            time semantic attribute using the same string with these two
            expressions:

                RawExpression(query="CAST(from_unixtime(self.time) AS DATE)")
                SemanticExpression(
                    query="CAST(from_unixtime(self.time) AS DATE)"
                )

            If we then wanted to create an attribute `month` it is likely
            easiest to use the semantic expression, but also possible to do this
            as a RawExpression. Here's how you'd do it in both cases:

                RawExpression(
                    query="timestamp_trunc('month',from_unixtime(self.time))"
                )
                SemanticExpression(query="timestamp_trunc('month', self.date)")
    """
        ),
    )
    column_name: str = Field(
        default="",
        description=textwrap.dedent(
            """
        The name of the column on the table, this _must_ be a column on the
        actual underlying table in the database. This is automatically converted
        into a RawExpression.
    """
        ),
    )

    def __hash__(self):
        return self.name.__hash__()

    @model_validator(mode="after")
    def process_dimension(self):
        if self.query and self.column_name:
            raise ValueError("Cannot specify both query and column_name for dimension")
        if not self.query and not self.column_name:
            self.column_name = self.name

        if self.query:
            if isinstance(self.query, str):
                # If the query is a string, we assume it's a semantic expression
                self.query = SemanticExpression(query=self.query)
            expression = self.query
        else:
            expression = RawExpression(
                query=exp.Column(
                    table=exp.to_identifier("self"),
                    this=exp.to_identifier(self.column_name),
                ).sql(dialect="duckdb")
            )

        match expression:
            case RawExpression():
                # check the columns in the expression and ensure they only reference `self`
                references: t.List[AttributePath] = []
                columns = list(expression.columns())
                for column in columns:
                    if column.table != "self":
                        raise ValueError(
                            f"Dimension {self.name} cannot reference another model. Found reference to {column.table}. Only use `self`"
                        )
            case SemanticExpression():
                # Ensure all references point to self
                columns = []
                references = expression.references(allow_self=True)
                for reference in references:
                    if not reference.is_self_reference():
                        raise ValueError(
                            f"Dimension {self.name} cannot reference another model. Found reference to {reference}. Only use `self`"
                        )
        self._columns = columns
        self._expression = expression

        return self

    @property
    def columns(self) -> t.List[exp.Column]:
        """Returns the actual columns that this dimension defines. The Model
        class collects these in order to determine the minimal compatible schema
        for that the underlying table must implement."""

        return list(self._columns)

    def bind(self, model: "Model") -> "BoundDimension":
        return BoundDimension(
            model=model,
            dimension=self,
        )

    def resolve(
        self,
        self_model_name: str,
        registry: Registry,
        traverser: AttributePathTraverser | None = None,
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the dimension."""

        return self._expression.resolve(
            traverser=traverser,
            self_model_name=self_model_name,
            registry=registry,
        )

    def references(self, allow_self: bool = False) -> t.List["AttributePath"]:
        """Returns the semantic references in the dimension expression"""
        if isinstance(self._expression, SemanticExpression):
            return self._expression.references(allow_self=allow_self)
        else:
            return []


class BoundDimension:
    model_config = ConfigDict(arbitrary_types_allowed=True)

    model: "Model"
    dimension: Dimension

    def __init__(self, model: "Model", dimension: Dimension):
        self.model = model
        self.dimension = dimension

    def resolve(
        self,
        registry: Registry,
        traverser: "AttributePathTraverser | None" = None,
    ) -> exp.Expression:
        """Returns the QueryComponent"""

        return self.dimension.resolve(
            traverser=traverser,
            self_model_name=self.model.name,
            registry=registry,
        )

    def references(self) -> t.List["AttributePath"]:
        """Returns the semantic references in the dimension expression"""
        return [
            ref.rewrite_self(self.model.name) for ref in self.dimension.references()
        ]


class RelationshipType(str, Enum):
    """The type of reference to use for the model.

    This is used to determine how to join the models together.
    """

    ONE_TO_ONE = "one_to_one"
    MANY_TO_ONE = "many_to_one"
    ONE_TO_MANY = "one_to_many"


class Relationship(BaseModel):
    """Defines a relationship between two models in the semantic layer.

    A Relationship represents how one model connects to another model through
    foreign key references. This enables the semantic layer to understand
    data relationships and generate appropriate JOIN clauses in SQL queries.

    Attributes:
        name: Optional name for the relationship. If not provided, defaults to ref_model.
        description: Optional human-readable description of the relationship.
        type: The type of relationship (ONE_TO_ONE, ONE_TO_MANY, MANY_TO_ONE).
        source_foreign_key: Column(s) in the source (current) model that reference the target model.
                           Can be a single column name or list of column names for composite keys.
        ref_model: Name of the target model being referenced.
        ref_key: Column(s) in the target model that is referenced.
                        Can be a single column name or list of column names for composite keys.
    """

    name: str = ""
    description: str = ""
    type: RelationshipType
    source_foreign_key: str | list[str]
    ref_model: str
    ref_key: str | list[str]

    def __hash__(self):
        return self.name.__hash__()

    @model_validator(mode="after")
    def process_relationship(self):
        if not self.name:
            # If no name is provided, use the model_ref as the name
            self.name = self.ref_model

        if isinstance(self.source_foreign_key, str):
            self.source_foreign_key = [self.source_foreign_key]
        if isinstance(self.ref_key, str):
            self.ref_key = [self.ref_key]

        assert len(self.source_foreign_key) == len(self.ref_key), (
            f"Source foreign key {self.source_foreign_key} and reference foreign key {self.ref_key} must have the same length"
        )

        self._expression = RawExpression(
            query=exp.Column(
                table=exp.to_identifier("self"),
                this=exp.to_identifier(self.source_foreign_key[0]),
            ).sql(dialect="trino")
        )

        return self

    def bind(self, model: "Model") -> "BoundRelationship":
        """Binds the relationship to a model and returns a BoundRelationship"""

        return BoundRelationship(
            model=model,
            relationship=self,
        )

    def inverse(self, model: "Model") -> "Relationship":
        """Returns the inverse relationship of this relationship.

        The inverse relationship is the relationship from the target model to the source model.
        """
        return Relationship(
            name=self.name,
            description=self.description,
            type=self.type,
            source_foreign_key=self.ref_key,
            ref_model=model.name,
            ref_key=self.source_foreign_key,
        )

    def references(self, model: "Model") -> t.List["AttributePath"]:
        """Returns the semantic references in the relationship"""
        return [AttributePath.from_string(f"{model.name}.{self.name}")]

    def resolve(
        self,
        self_model_name: str,
        registry: Registry,
        traverser: AttributePathTraverser | None = None,
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the relationship."""

        return self._expression.resolve(
            traverser=traverser,
            self_model_name=self_model_name,
            registry=registry,
        )


class BoundRelationship:
    model: "Model"
    relationship: Relationship

    def __init__(self, *, model: "Model", relationship: Relationship):
        self.model = model
        self.relationship = relationship

    def inverse(self, model: "Model") -> "BoundRelationship":
        return BoundRelationship(
            model=model, relationship=self.relationship.inverse(model)
        )

    def resolve(
        self,
        registry: Registry,
        traverser: "AttributePathTraverser | None" = None,
    ) -> exp.Expression:
        """Returns the QueryComponent"""

        return self.relationship.resolve(
            traverser=traverser,
            self_model_name=self.model.name,
            registry=registry,
        )


class ViewConnector(BaseModel):
    def primary_key(self, model_name) -> t.Optional[str | exp.Expression]:
        return None

    def time_column(self, model_name) -> t.Optional[str]:
        return None

    def dimensions(self, model_name) -> t.List[Dimension]:
        return []

    def measures(self, model_name) -> t.List[Measure]:
        return []


class View(BaseModel):
    """A generic set of dimensions and measures that can be queried together."""

    name: str
    description: str = ""
    connector: ViewConnector


class Model(BaseModel):
    """A model is a semantic layer representation of a table in the warehouse as
    a set of dimensions, measures, and relationships."""

    name: str
    description: str = ""
    table: str

    primary_key: str | t.List[str] = ""
    time_column: t.Optional[str] = None

    # These are additional interfaces to the model
    views: t.List[View | str] = Field(default_factory=lambda: [])

    dimensions: t.List[Dimension] = Field(default_factory=lambda: [])
    measures: t.List[Measure] = Field(default_factory=lambda: [])
    relationships: t.List[Relationship] = Field(default_factory=lambda: [])

    @model_validator(mode="after")
    def build_attribute_lookup(self):
        """Builds a lookup of the references in the model"""
        attributes: dict[str, BoundDimension | BoundMeasure | BoundRelationship] = {}
        known_columns: t.Set[str] = set()

        # Topologically sort all of the dimensions, measures, and relationships in the model
        attribute_graph: t.Dict[str, t.Set[str]] = {}
        attribute_references: t.Dict[str, t.List[AttributePath]] = {}

        for dimension in self.dimensions:
            if dimension.name in attribute_graph:
                raise ValueError(
                    f"Dimension {dimension.name} already exists in model {self.name}"
                )
            references = [
                ref.rewrite_self(self.name)
                for ref in dimension.references(allow_self=True)
            ]

            known_columns.update(
                set([exp_to_str(col.this) for col in dimension.columns]),
            )
            attribute_graph[dimension.name] = set()
            attribute_references[dimension.name] = references

            for reference in references:
                if reference.is_relationship() or reference.base_model != self.name:
                    raise ValueError("Cannot reference a relationship in a dimension")
                # Attribute/Column name
                attribute_name = exp_to_str(reference.final_attribute().this)

                if attribute_name == dimension.name:
                    # If the attribute name is the same as the dimension name, we can skip it.
                    # For a dimension this just means that the dimension is a column in the table
                    continue

                attribute_graph[dimension.name].add(
                    exp_to_str(reference.final_attribute().this)
                )

            attributes[dimension.name] = dimension.bind(self)

        # Check for cycles in the dimensions
        dimensions_sorter = TopologicalSorter(attribute_graph)
        dimensions_sorter.prepare()

        # Check for cycles in the measures and dimensions
        # For now this only checks for cycles in the current model.
        for measure in self.measures:
            # Ensure we don't have duplicate names
            if measure.name in attributes or measure.name in attribute_graph:
                raise ValueError(
                    f"Measure {measure.name} already exists in model {self.name}"
                )
            references = [
                ref.rewrite_self(self.name)
                for ref in measure.references(allow_self=True)
            ]
            attribute_graph[measure.name] = set()
            attribute_references[measure.name] = references

            known_columns.update(
                set([exp_to_str(col.this) for col in measure.columns]),
            )

            for reference in references:
                reference = reference.rewrite_self(self.name)
                if reference.is_relationship() or reference.base_model != self.name:
                    # Relationships are allowed in measures and must be validated at the
                    # registry level
                    continue
                attribute_graph[measure.name].add(
                    exp_to_str(reference.final_attribute().this)
                )

            attributes[measure.name] = measure.bind(self)

        # Ensure we have no cycles in the attribute graph
        attribute_sorter = TopologicalSorter(attribute_graph)
        attribute_sorter.prepare()

        self._model_ref_lookup_by_model_name: dict[str, list[Relationship]] = {}
        for reference in self.relationships:
            if reference.name in attributes:
                raise ValueError(
                    f"{reference.name} already exists in model {self.name}. Must be unique"
                )

            attributes[reference.name] = reference.bind(self)
            attribute_references[reference.name] = []

            if isinstance(reference.source_foreign_key, str):
                known_columns.add(reference.source_foreign_key)
            else:
                known_columns.update(reference.source_foreign_key)

            if reference.ref_model not in self._model_ref_lookup_by_model_name:
                self._model_ref_lookup_by_model_name[reference.ref_model] = []
            self._model_ref_lookup_by_model_name[reference.ref_model].append(reference)

        self._all_attributes = attributes

        if isinstance(self.primary_key, str):
            self._primary_key_expression = coerce_to_tables_column(self.primary_key)
        elif isinstance(self.primary_key, list):
            self._primary_key_expression = hash_expressions(
                *[coerce_to_tables_column(pk) for pk in self.primary_key]
            )

        self._known_columns = known_columns
        self._attribute_references = attribute_references
        return self

    @property
    def known_columns(self) -> t.Set[str]:
        """Returns the known columns in the model

        This is useful for validating that the model is valid for a given sql
        connection. This set of columns should comprise the schema of the table.
        It isn't required to be exhaustive, but at a minimum the connection
        should have the same columns as the known columns.
        """
        return self._known_columns

    def primary_key_expression(self, table_alias: str) -> exp.Expression:
        """Returns the primary key expression for the model given the table alias."""

        def transform_columns(node: exp.Expression):
            if isinstance(node, exp.Column):
                node_table = exp_to_str(node.table)
                if node_table != "self":
                    # Do a check here but this is pretty fatal and should have
                    # been caught at validation time
                    raise ValueError(
                        f"Primary key column {node_table}.{node.this} must be a self reference"
                    )
                return exp.Column(
                    table=exp.to_identifier(table_alias),
                    this=node.this,
                )
            return node

        primary_key_expression = self._primary_key_expression.transform(
            transform_columns
        )
        return primary_key_expression

    @property
    def attribute_names(self) -> t.List[str]:
        """Returns the names of all attributes in the model"""
        return list(self._all_attributes.keys())

    def get_dimension(self, name: str) -> BoundDimension:
        """Returns the dimension with the given name"""
        dimension = self.get_attribute(name)
        if not isinstance(dimension, BoundDimension):
            raise ValueError(f"{name} is not a dimension in model {self.name}")
        return dimension

    def get_relationship(self, name: str) -> BoundRelationship:
        """Returns the relationship with the given name"""
        relationship = self.get_attribute(name)
        if not isinstance(relationship, BoundRelationship):
            raise ValueError(f"{name} is not a relationship in model {self.name}")
        return relationship

    def find_relationship(
        self, *, model_ref: "Model", name: str = "", model_ref_lookup: bool = False
    ) -> "BoundRelationship":
        """Returns the reference with the given name or model_ref"""
        model_ref_name = model_ref.name
        if name:
            for reference in self.relationships:
                if reference.name == name:
                    if reference.ref_model != model_ref_name:
                        raise ValueError(
                            f"Reference {name} does not match model_ref {model_ref_name}"
                        )
                    return self.get_relationship(name)

        for reference in self.relationships:
            references = self._model_ref_lookup_by_model_name[reference.ref_model]
            if len(references) > 1:
                raise ModelHasAmbiguousJoinPath(
                    f"Reference {model_ref_name} is ambiguous in model {self.name}"
                )
            if reference.ref_model == model_ref_name:
                relationship = self._all_attributes[reference.name]
                if not isinstance(relationship, BoundRelationship):
                    raise ValueError(
                        f"{reference.name} is not a reference in model {self.name}"
                    )
                return relationship

        # Check if the relationship is defined on the other model
        if not model_ref_lookup:
            relationship = model_ref.find_relationship(
                model_ref=self, model_ref_lookup=True
            )
            return relationship.inverse(model_ref)

        raise ValueError(f"Reference {model_ref_name} not found in model {self.name}")

    @property
    def table_exp(self):
        """Returns the table for the model"""
        return exp.to_table(self.table)

    def get_attribute(
        self, name: str
    ) -> BoundDimension | BoundMeasure | BoundRelationship:
        """Resolves the attribute to a column in the model"""
        if name not in self._all_attributes:
            raise ValueError(f"Attribute {name} not found in model {self.name}")
        return self._all_attributes[name]

    def attributes(self):
        """Returns all attributes in the model"""
        return self._all_attributes.values()

    def describe(self):
        """Returns a description of the model used for an LLM prompt"""

        description = f"## Model: {self.name}\n\n"
        description += f"{self.description}\n"
        description += "\n### Dimensions:\n"
        for dimension in self.dimensions:
            description += (
                f"- `{self.name}.{dimension.name}`: {dimension.description}\n"
            )

        description += "\n### Measures:\n"
        for measure in self.measures:
            description += f"- `{self.name}.{measure.name}`: {measure.description}\n"

        description += "\n### Relationships:\n"
        for relationship in self.relationships:
            description += (
                f"- `{self.name}.{relationship.name}`: "
                f"{relationship.description}. This is a {relationship.type.value} relationship\n"
            )
        ambiguous_rels_descriptions = ""
        include_ambiguous = False
        for (
            model_ref,
            model_relationships,
        ) in self._model_ref_lookup_by_model_name.items():
            if len(model_relationships) > 1:
                include_ambiguous = True
                ambiguous_rels_descriptions += (
                    f"- `{model_ref}`:"
                    f" Any relationships to this model or anything derived from it"
                    f" must use the `->` through syntax to prevent ambiguous joins in the query\n"
                )
        if include_ambiguous:
            description += "\n### Ambiguous Relationships:\n"
            description += ambiguous_rels_descriptions
        description += "\n"
        return description

    def get_attribute_references(self, name: str) -> t.List[AttributePath]:
        """Returns the attribute references for the given attribute name"""
        if name not in self._attribute_references:
            raise ValueError(f"Attribute {name} not found in model {self.name}")
        return self._attribute_references[name]


class RollingWindow(BaseModel):
    window: int
    unit: str


class ModelHasNoJoinPath(Exception):
    pass


class ModelHasAmbiguousJoinPath(Exception):
    pass


class Join(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    table: str
    via: str | None = None


class AttributePath(BaseModel):
    path: list[str]

    @model_validator(mode="after")
    def process_ref(self):
        self._columns = [exp.to_column(r.strip()) for r in self.path]

        # Write out normalized path
        self.path = [
            f"{exp_to_str(column.table)}.{exp_to_str(column.this)}"
            for column in self._columns
        ]
        return self

    @classmethod
    def from_string(cls, path: str) -> "AttributePath":
        return cls(path=path.split("->"))

    def traverser(
        self,
        initial_value: str = "",
        parent_traverser: "AttributePathTraverser | None" = None,
    ):
        return AttributePathTraverser.from_reference(
            self,
            initial_value=initial_value,
            parent_traverser=parent_traverser,
        )

    @property
    def base_model(self):
        return self._columns[0].table

    @property
    def columns(self):
        return self._columns

    def is_valid_for_registry(self, registry: Registry):
        """Checks if the reference is valid for the given registry"""
        return self.traverser().is_valid_for_registry(registry)

    def __str__(self) -> str:
        """Returns the string representation of the reference"""
        columns_as_strs = [
            f"{exp_to_str(column.table)}.{exp_to_str(column.this)}"
            for column in self._columns
        ]
        return "->".join(columns_as_strs)

    def to_select_alias(self) -> str:
        """Returns the string representation of the reference as a select alias"""
        columns_as_strs = [
            f"{exp_to_str(column.table)}_{exp_to_str(column.this)}"
            for column in self._columns
        ]
        return "__".join(columns_as_strs)

    def __eq__(self, other: object) -> bool:
        """Checks if the reference is equal to another reference"""
        if not isinstance(other, AttributePath):
            return False
        return self.path == other.path

    def final_attribute(self) -> exp.Column:
        """Returns the final column in the reference"""
        return self._columns[-1]

    def is_relationship(self) -> bool:
        """Checks if the reference is a relationship"""
        return len(self._columns) > 1

    def is_self_reference(self) -> bool:
        """Checks if the reference is a self reference"""
        has_self = False
        for column in self._columns:
            if exp_to_str(column.table) == "self":
                has_self = True
        return has_self

    def rewrite_self(self, self_model_name: str) -> "AttributePath":
        """Rewrites the self reference to the given model name"""
        new_path: list[str] = []
        for column in self._columns:
            if exp_to_str(column.table) == "self":
                new_path.append(f"{self_model_name}.{exp_to_str(column.this)}")
            else:
                new_path.append(f"{exp_to_str(column.table)}.{exp_to_str(column.this)}")
        return AttributePath(path=new_path)


class AttributeTransformerResult(BaseModel):
    """Internal result only used for the attribute reference transformer"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    node: exp.Expression
    references: list[AttributePath]


def is_expression_attribute_reference(node: exp.Expression) -> bool:
    if isinstance(node, exp.Column):
        return True

    if not isinstance(node, exp.JSONExtract):
        return False

    base = node.this
    to = node.expression
    if not isinstance(base, (exp.Column, exp.Dot)):
        return False

    if not isinstance(to, (exp.Column, exp.JSONExtract)):
        return False

    if isinstance(to, exp.JSONExtract):
        return is_expression_attribute_reference(to)
    return True


class QueryComponent(t.Protocol):
    """The most basic representation of a part of a query.

    This is used to represent dimensions, measures, and filters in a query.

    Each of those parts of a query should have the following properties:

    * An expression that can be used to generate the SQL
    * 1 or more references related to the expression. We need to know all of
      these to understand how to generate the right joins
    * If the expression is an aggregate function, we need to know if it is
        additive or not. This is used to determine if we need to use ctes or
        not. If the expression is not an aggregate function, we can group by
        this expression.

    The `original` field is used to store the original attribute path or
    SemanticExpression that was used to create this query component.

    The `expression` field is a sqlglot expression that has not been fully
    resolved for the current query. Resolving replaces any column references or
    semantic attributes to the actual columns in the query.
    """

    def is_aggregate(self, registry: "Registry") -> bool:
        """Returns whether the query component is an aggregate function."""
        ...

    def resolve(
        self, registry: "Registry", traverser: "AttributePathTraverser | None" = None
    ) -> exp.Expression:
        """Returns a resolved SQL expression for the query component.

        This should resolve any references to the actual columns in the query.
        """
        ...

    def references(self) -> t.List[AttributePath]:
        """Returns the semantic references in the query component."""
        ...

    def alias(self) -> str:
        """Returns the alias for the query component. If the component cannot be
        aliased, this should return an empty string.

        This is used to generate the SQL alias for the query component.
        """
        ...


class AttributePathTraverser:
    """AttributeReferences are used to reference a dimension or metric in a
    model.

    Because some models may reference another model more than once. The way this
    would be distinguished would be through different model relationship
    attributes. Path traversal, tracks the joins through these attributes.

    The following table shows how a traverser works, given a path
    `a.b->c.d->e.f`:

    | Index             | 0      | 1               | 2                        |
    |-------------------|--------|-----------------|--------------------------|
    | Through Column    |        | a.b             | c.d                      |
    | Alias for table g | md5(g) | md5(md5(a.b)+g) | md5(md5(md5(a.b)+c.d)+g) |

    This allows us to ensure that we don't duplicate joins by ensuring that each
    level through a given column reference uses the same table alias.
    """

    @classmethod
    def from_root(cls, initial_value: str = ""):
        """Creates a traverser from the root of the attribute path"""
        return cls(reference=AttributePath(path=[]), initial_value=initial_value)

    @classmethod
    def from_reference(
        cls,
        reference: AttributePath,
        initial_value="",
        parent_traverser: "AttributePathTraverser | None" = None,
    ):
        """Creates a traverser from an AttributePath reference"""
        if parent_traverser:
            reference = parent_traverser.create_scoped_reference(reference)

        return cls(reference=reference, initial_value=initial_value)

    def __init__(self, reference: AttributePath, initial_value=""):
        self.reference = reference
        self.index = 0
        self.alias_stack: list[str] = [initial_value]

    def next(self):
        if self.index >= len(self.reference.columns) - 1:
            return False

        self.alias_stack.append(
            self._generate_alias(
                self.alias_stack[-1],
                self.current_model_name,
                self.current_attribute_name,
            )
        )
        self.index += 1
        return True

    def prev(self):
        if self.index == 0:
            return False
        self.alias_stack.pop()
        self.index -= 1
        return True

    def _generate_alias(self, prev_alias: str, current_table: str, column: str = ""):
        m = hashlib.md5()
        m.update(prev_alias.encode("utf-8"))
        m.update(current_table.encode("utf-8"))
        if column:
            m.update(column.encode("utf-8"))
        return f"{current_table}_{m.hexdigest()[:8]}"

    def alias(self, model_name: str):
        """Creates an alias at the given traverser position"""
        return self._generate_alias(self.alias_stack[-1], model_name)

    def copy(self):
        """Creates a copy of the traverser"""
        traverser = AttributePathTraverser(
            reference=self.reference,
            initial_value=self.alias_stack[-1],
        )
        traverser.index = self.index
        traverser.alias_stack = self.alias_stack[:]
        return traverser

    def create_scoped_reference(self, reference: AttributePath) -> AttributePath:
        """Creates a scoped reference for the given model and column"""
        if self.index == 0:
            return reference
        else:
            new_ref = self.reference.path[: self.index]
            new_ref.extend(reference.path)
            return AttributePath(path=new_ref)

    @property
    def current_table_alias(self):
        return self.alias(self.current_model_name)

    @property
    def current_model_name(self):
        return self.reference.columns[self.index].table

    @property
    def current_attribute_name(self):
        return self.reference.columns[self.index].name

    def is_valid_for_registry(self, registry: Registry):
        """Checks if the reference is valid for the given registry"""

        # The final attribute must be a dimension or metric so we only iterate
        # through the relationships if we are not at the end of the reference
        if len(self.reference.columns) > 1:
            current_model = registry.get_model(self.current_model_name)
            current_attribute = current_model.get_attribute(self.current_attribute_name)
            # Only relationships can be traversed any other kind of attribute is
            # invalid
            if not isinstance(current_attribute, BoundRelationship):
                raise InvalidAttributeReferenceError(self.reference)
            while self.next():
                pass
        current_model = registry.get_model(self.current_model_name)
        current_attribute = current_model.get_attribute(self.current_attribute_name)
        if isinstance(current_attribute, BoundRelationship):
            raise InvalidAttributeReferenceError(self.reference)
        return True

    def to_end(self):
        """Moves the traverser to the end of the reference"""
        while self.next():
            pass
        return self


class SemanticQuery(BaseModel):
    """Used for serialization of semantic queries"""

    name: str = Field(
        default="", description="Optional name used to identify the query"
    )

    description: str = Field(
        default="", description="A description of the query's intent and purpose."
    )

    selects: t.List[str] = Field(
        description=textwrap.dedent(
            """
        A list of model attributes to select in the query. These should only be
        valid dimensions or measures from the available models in the registry.
        Joins are automatically handled by the query builder based on the
        provided attributes and the relationships defined in the models. If any
        ambiguous relationships between models should use the `->` syntax to
        specify the path through the relationships.
    """
        )
    )
    filters: t.List[str] = Field(
        description=textwrap.dedent(
            """
        A list of filters to apply to the query. These should be valid
        expressions that can be used to filter the results. The expressions can
        reference model attributes just like the columns. If any ambiguous
        relationships between models should use the `->` syntax to specify the
        path through the relationships.
    """
        ),
        default_factory=lambda: [],
    )

    limit: int = Field(
        default=0,
        description=textwrap.dedent(
            """
            The maximum number of rows to return in the query. If set to 0, no
            limit is applied.
        """
        ),
    )



from sqlglot import exp

from .query import QueryBuilder
from .testing import setup_registry


def test_query_builder_with_dimensions():
    registry = setup_registry()

    query = QueryBuilder(registry)
    query.select("int_events__github.month AS event_month")
    query.select("int_events__github.event_source AS event_source")
    query.select("int_events__github.from->artifacts.artifact_name")
    query.select("int_events__github.to->collections.collection_name")
    query.where("int_events__github.month = DATE '2023-01-01'")

    query_exp = query.build()

    joins = list(query_exp.find_all(exp.Join))

    assert len(joins) == 4

    tables = [j.this for j in joins]
    tables_count = {}
    for table in tables:
        table_count = tables_count.get(table.name, 0)
        tables_count[table.name] = table_count + 1

    assert tables_count["artifacts_v1"] == 2
    assert tables_count["collections_v1"] == 1


def test_query_builder_with_metrics():
    registry = setup_registry()

    query = QueryBuilder(registry)
    query.select("collections.collection_name")
    query.select("projects.count AS project_count")

    query_exp = query.build()

    joins = list(query_exp.find_all(exp.From)) + list(query_exp.find_all(exp.Join))

    assert len(joins) == 3

    tables = [j.this for j in joins]
    tables_count = {}
    for table in tables:
        table_count = tables_count.get(table.name, 0)
        tables_count[table.name] = table_count + 1

    assert tables_count["projects_v1"] == 1
    assert tables_count["collections_v1"] == 1


def test_query_builder_with_metric_and_filtered_dimension():
    registry = setup_registry()

    query = QueryBuilder(registry)
    query.select("projects.count")
    query.where("collections.collection_name = 'optimism'")

    query_exp = query.build()

    assert query_exp.expressions is not None
    assert len(query_exp.expressions) == 1
    alias = query_exp.expressions[0].alias
    assert alias == "projects_count"
    count_exp = query_exp.expressions[0].this
    assert isinstance(count_exp, exp.Count)
    column_exp = count_exp.this
    assert isinstance(column_exp, exp.Column)
    column_name_identifier = column_exp.this
    assert isinstance(column_name_identifier, exp.Identifier)
    assert column_name_identifier.this == "project_id"


def test_query_builder_to_model():
    registry = setup_registry()

    query = QueryBuilder(registry)
    query.select("int_events__github.month AS event_month")
    query.select("int_events__github.event_type AS event_type")
    query.select("int_events__github.event_source")
    query.select("int_events__github.from->artifacts.artifact_name")
    query.select("int_events__github.count")
    query.select("int_events__github.from AS artifact_id")

    # Attempt to convert the query to a model
    model = query.to_model("test_model")

    model.dimensions.sort(key=lambda d: d.name)

    assert model.name == "test_model"
    assert model.table == "test_model"
    assert len(model.dimensions) == 5
    assert model.dimensions[0].name == "event_month"
    assert model.dimensions[1].name == "event_type"
    assert model.dimensions[2].name == "int_events__github_count"
    assert model.dimensions[3].name == "int_events__github_event_source"
    assert (
        model.dimensions[4].name == "int_events__github_from__artifacts_artifact_name"
    )
    assert len(model.relationships) == 1
    assert model.relationships[0].name == "artifact_id"


def test_query_builder_with_cte():
    registry = setup_registry()

    # Create the first query and convert it to a CTE
    base_query = QueryBuilder(registry)
    base_query.select("int_events__github.month AS event_month")
    base_query.select("int_events__github.event_source AS event_source")
    base_query.select("int_events__github.to")

    # Register the CTE
    cte_name = "github_push_events"

    # Create a new query that references the CTE
    new_query = QueryBuilder(registry)
    new_query.cte(cte_name, base_query)
    new_query.select(f"{cte_name}.event_month")
    new_query.select(f"{cte_name}.event_source")
    new_query.where(f"{cte_name}.event_month = DATE '2023-01-01'")

    # Build the final query
    query_exp = new_query.build()

    # Assertions to validate the query structure
    assert query_exp is not None
    assert len(list(query_exp.find_all(exp.From))) == 2  # Ensure the CTE is referenced
    assert (
        len(list(query_exp.find_all(exp.Join))) == 0
    )  # No joins expected in this case



from ..definition import Registry
from .entities import register_entities
from .events import register_events
from .metrics import register_metrics


def register_oso_models(registry: Registry, catalog_name: str = "iceberg"):
    """Register OSO models based on the warehouse/oso_sqlmesh/models/marts structure.

    This function registers all the core entities, metrics, and events from the
    OSO data warehouse.
    """
    register_entities(registry, catalog_name)
    register_metrics(registry, catalog_name)
    register_events(registry, catalog_name)



import textwrap

from ..definition import (
    Dimension,
    Measure,
    Model,
    Registry,
    Relationship,
    RelationshipType,
)


def register_events(registry: Registry, catalog_name: str = "iceberg"):
    # ============================================"
    # EVENTS
    # ============================================"

    registry.register(
        Model(
            name="event_types",
            table=f"{catalog_name}.oso.event_types_v1",
            description=textwrap.dedent(
                """
                The event_types_v1 table contains distinct event types from various sources, 
                such as GitHub, blockchain, and funding events. This table is used to enumerate 
                different types of events that occur across the open source ecosystem.
                """
            ),
            dimensions=[
                Dimension(
                    name="event_type",
                    description="The unique identifier for the event type",
                    column_name="event_type",
                ),
            ],
            primary_key="event_type",
            measures=[
                Measure(
                    name="count",
                    description="The number of distinct event types",
                    query="COUNT(*)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="int_events__github",
            table=f"{catalog_name}.oso.int_events__github",
            description=textwrap.dedent(
                """
                GitHub events aggregated daily. This includes all types of GitHub activity
                such as commits, pull requests, issues, stars, forks, and more.
                """
            ),
            dimensions=[
                Dimension(
                    name="time",
                    description="The timestamp when the event occurred",
                    column_name="time",
                ),
                Dimension(
                    name="event_type",
                    description="The type of the GitHub event",
                    column_name="event_type",
                ),
                Dimension(
                    name="event_source",
                    description="The source of the event (always GITHUB)",
                    column_name="event_source",
                ),
                Dimension(
                    name="event_source_id",
                    description="The source ID of the event",
                    column_name="event_source_id",
                ),
                Dimension(
                    name="to_artifact_id",
                    description="The ID of the artifact to which the event occurred",
                    column_name="to_artifact_id",
                ),
                Dimension(
                    name="from_artifact_id",
                    description="The ID of the artifact from which the event occurred",
                    column_name="from_artifact_id",
                ),
                Dimension(
                    name="to_artifact_name",
                    description="The name of the target artifact",
                    column_name="to_artifact_name",
                ),
                Dimension(
                    name="to_artifact_namespace",
                    description="The namespace of the target artifact",
                    column_name="to_artifact_namespace",
                ),
                Dimension(
                    name="to_artifact_source_id",
                    description="The source ID of the target artifact",
                    column_name="to_artifact_source_id",
                ),
                Dimension(
                    name="from_artifact_name",
                    description="The name of the source artifact",
                    column_name="from_artifact_name",
                ),
                Dimension(
                    name="from_artifact_namespace",
                    description="The namespace of the source artifact",
                    column_name="from_artifact_namespace",
                ),
                Dimension(
                    name="from_artifact_source_id",
                    description="The source ID of the source artifact",
                    column_name="from_artifact_source_id",
                ),
                Dimension(
                    name="amount",
                    description="The amount or count of the event",
                    column_name="amount",
                ),
                Dimension(
                    name="month",
                    description="The month the event occurred",
                    query="DATE_TRUNC('month', self.time)",
                ),
                Dimension(
                    name="year",
                    description="The year the event occurred",
                    query="DATE_TRUNC('year', self.time)",
                ),
                Dimension(
                    name="date",
                    description="The date the event occurred",
                    query="DATE_TRUNC('day', self.time)",
                ),
            ],
            measures=[
                Measure(
                    name="count",
                    description="The number of events",
                    query="COUNT(*)",
                ),
                Measure(
                    name="total_amount",
                    description="The total amount of events",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg_amount",
                    description="The average amount of events",
                    query="AVG(self.amount)",
                ),
            ],
            time_column="time",
            relationships=[
                Relationship(
                    name="from",
                    description="The artifact from which the event occurred",
                    source_foreign_key="from_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    name="to",
                    description="The artifact to which the event occurred",
                    source_foreign_key="to_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="int_events_daily__blockchain",
            table=f"{catalog_name}.oso.int_events_daily__blockchain",
            description=textwrap.dedent(
                """
                Blockchain events aggregated daily. This includes various blockchain activities
                such as transactions, token transfers, contract interactions, and more.
                """
            ),
            dimensions=[
                Dimension(
                    name="bucket_day",
                    description="The day the event occurred",
                    column_name="bucket_day",
                ),
                Dimension(
                    name="event_type",
                    description="The type of the blockchain event",
                    column_name="event_type",
                ),
                Dimension(
                    name="event_source",
                    description="The source blockchain",
                    column_name="event_source",
                ),
                Dimension(
                    name="from_artifact_id",
                    description="The ID of the artifact from which the event occurred",
                    column_name="from_artifact_id",
                ),
                Dimension(
                    name="to_artifact_id",
                    description="The ID of the artifact to which the event occurred",
                    column_name="to_artifact_id",
                ),
                Dimension(
                    name="amount",
                    description="The amount or count of the event",
                    column_name="amount",
                ),
                Dimension(
                    name="count",
                    description="The count of events",
                    column_name="count",
                ),
                Dimension(
                    name="month",
                    description="The month the event occurred",
                    query="DATE_TRUNC('month', self.bucket_day)",
                ),
                Dimension(
                    name="year",
                    description="The year the event occurred",
                    query="DATE_TRUNC('year', self.bucket_day)",
                ),
            ],
            measures=[
                Measure(
                    name="record_count",
                    description="The number of event records",
                    query="COUNT(*)",
                ),
                Measure(
                    name="total_amount",
                    description="The total amount of events",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg_amount",
                    description="The average amount of events",
                    query="AVG(self.amount)",
                ),
                Measure(
                    name="total_count",
                    description="The total count from the count column",
                    query="SUM(self.count)",
                ),
            ],
            time_column="bucket_day",
            relationships=[
                Relationship(
                    name="from",
                    description="The artifact from which the event occurred",
                    source_foreign_key="from_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    name="to",
                    description="The artifact to which the event occurred",
                    source_foreign_key="to_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="int_events_daily__funding",
            table=f"{catalog_name}.oso.int_events_daily__funding",
            description=textwrap.dedent(
                """
                Funding events aggregated daily. This includes funding activities from various
                sources such as Gitcoin, Open Collective, and other funding platforms.
                """
            ),
            dimensions=[
                Dimension(
                    name="bucket_day",
                    description="The day the event occurred",
                    column_name="bucket_day",
                ),
                Dimension(
                    name="event_type",
                    description="The type of the funding event",
                    column_name="event_type",
                ),
                Dimension(
                    name="event_source",
                    description="The source of the funding",
                    column_name="event_source",
                ),
                Dimension(
                    name="from_artifact_id",
                    description="The ID of the artifact from which the event occurred",
                    column_name="from_artifact_id",
                ),
                Dimension(
                    name="to_artifact_id",
                    description="The ID of the artifact to which the event occurred",
                    column_name="to_artifact_id",
                ),
                Dimension(
                    name="amount",
                    description="The funding amount",
                    column_name="amount",
                ),
                Dimension(
                    name="month",
                    description="The month the event occurred",
                    query="DATE_TRUNC('month', self.bucket_day)",
                ),
                Dimension(
                    name="year",
                    description="The year the event occurred",
                    query="DATE_TRUNC('year', self.bucket_day)",
                ),
            ],
            measures=[
                Measure(
                    name="count",
                    description="The number of events",
                    query="COUNT(*)",
                ),
                Measure(
                    name="total_amount",
                    description="The total amount of events",
                    query="SUM(self.amount)",
                ),
                Measure(
                    name="avg_amount",
                    description="The average amount of events",
                    query="AVG(self.amount)",
                ),
            ],
            time_column="bucket_day",
            relationships=[
                Relationship(
                    name="from",
                    description="The artifact from which the event occurred",
                    source_foreign_key="from_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    name="to",
                    description="The artifact to which the event occurred",
                    source_foreign_key="to_artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )



import logging
import typing as t
from copy import deepcopy

from sqlglot import exp

from .definition import (
    AttributePath,
    AttributePathTraverser,
    BoundMeasure,
    BoundRelationship,
    Dimension,
    Filter,
    JoinTree,
    Model,
    QueryComponent,
    QueryRegistry,
    Registry,
    Relationship,
    Select,
    SemanticExpression,
)
from .utils import exp_to_str

logger = logging.getLogger(__name__)


class QueryBuilder(QueryRegistry):
    def __init__(self, registry: Registry):
        self._registry = deepcopy(registry)
        self._ctes: list[tuple[str, QueryBuilder]] = []
        self._select_refs: list[AttributePath] = []
        self._references: list[AttributePath] = []
        self._root_model: Model | None = None

        self._select_parts: list[QueryComponent] = []
        self._select_aliases: list[str] = []
        self._filter_parts: list[QueryComponent] = []

        self._limit = 0

    def _add_reference(self, reference: AttributePath):
        """Adds an attribute reference to the query

        Every reference adds 0 or more joins to the query
        """
        self._references.append(reference)

        return self

    def select(self, *selects: str):
        """Add a model attribute to the select clause"""
        for select in selects:
            select_expr = Select(query=select)

            alias = select_expr.alias()

            references = select_expr.references()
            reference = references[0]

            # validate the select by checking the attribute references

            resolved_references = self._registry.expand_reference(reference)

            for resolved_reference in resolved_references:
                if resolved_reference not in self._references:
                    self._add_reference(resolved_reference)
            self._select_parts.append(select_expr)
            self._select_aliases.append(alias)
        return self

    def where(self, *filters: str | SemanticExpression):
        """Add a filter to the query"""
        for filter in filters:
            filter_expr = Filter(query=filter)

            references = filter_expr.references()

            for reference in references:
                if not reference.is_valid_for_registry(self._registry):
                    raise ValueError(f"Invalid reference {reference} for registry")
                # Resolve the reference to the actual attribute
                resolved_references = self._registry.expand_reference(reference)
                for resolved_reference in resolved_references:
                    if resolved_reference not in self._references:
                        self._add_reference(resolved_reference)

            self._filter_parts.append(filter_expr)
        return self

    def limit(self, limit: int):
        """Add a limit to the query"""
        self._limit = limit
        return self

    def cte(self, name: str, query: t.Self):
        model = query.to_model(name)
        self._registry.register(model)
        self._ctes.append((name, query))
        return self

    def to_model(self, name: str):
        relationships: set[Relationship] = set()
        dimensions: set[Dimension] = set()
        for select_part, alias in zip(self._select_parts, self._select_aliases):
            references = select_part.references()
            reference = references[0]
            final_column = reference.final_attribute()
            model = self._registry.get_model(exp_to_str(final_column.table))
            attribute = model.get_attribute(exp_to_str(final_column.this))
            if isinstance(attribute, BoundRelationship):
                relationships.add(
                    Relationship(
                        name=alias,
                        type=attribute.relationship.type,
                        source_foreign_key=alias,
                        ref_model=attribute.relationship.ref_model,
                        ref_key=attribute.relationship.ref_key,
                    )
                )
            elif isinstance(attribute, BoundMeasure):
                dimensions.add(
                    Dimension(
                        name=alias,
                        description=attribute.measure.description,
                        column_name=alias,
                    )
                )
            else:
                dimensions.add(
                    Dimension(
                        name=alias,
                        description=attribute.dimension.description,
                        query="",
                        column_name=alias,
                    )
                )
                # Sometimes the relationship is also a dimension
                for relationship in model.relationships:
                    if relationship.source_foreign_key[0] == exp_to_str(
                        final_column.this
                    ):
                        relationships.add(
                            Relationship(
                                name=f"{relationship.ref_model}_by_{alias}",
                                type=relationship.type,
                                source_foreign_key=alias,
                                ref_model=relationship.ref_model,
                                ref_key=relationship.ref_key,
                            )
                        )

        return Model(
            name=name,
            table=name,
            dimensions=[dimension for dimension in dimensions],
            relationships=[relationship for relationship in relationships],
        )

    def build(self):
        """Render a select query"""
        join_tree = self._registry.dag.find_best_join_tree(self._references)
        self._root_model = self._registry.get_model(join_tree.root)

        # Turn references into actual expressions
        select_parts = self._select_parts
        select_expressions: list[exp.Expression] = []
        group_by_expressions: list[str] = []

        for i in range(len(select_parts)):
            part = select_parts[i]
            resolved = part.resolve(self._registry)
            alias = self._select_aliases[i]
            select_expressions.append(resolved.as_(alias))

            if not part.is_aggregate(self._registry):
                group_by_expressions.append(str(i + 1))

        # Establish base query
        query = exp.select(*select_expressions)

        base_model = self._root_model
        base_table = base_model.table_exp.as_(
            AttributePathTraverser.from_root().alias(base_model.name)
        )  # Use an empty path to get the root model alias
        query = query.from_(base_table)

        # Add joins
        joiner = QueryJoiner(query, base_model, join_tree, self._registry)
        for ref in sorted(
            self._references, key=lambda r: join_tree.depths[r.base_model]
        ):
            joiner.join_reference(ref)

        query = joiner.joined_query

        # Add filters
        for part in self._filter_parts:
            part_expression = part.resolve(self._registry)

            query = query.where(part_expression)

        # Apply appropriate group by
        if group_by_expressions:
            query = query.group_by(*group_by_expressions)

        if self._limit:
            query = query.limit(self._limit)

        # Add CTEs
        for name, cte_query in self._ctes:
            cte_exp = cte_query.build()
            query = query.with_(name, as_=cte_exp)

        return query


class QueryJoiner:
    def __init__(
        self,
        select: exp.Select,
        base_model: Model,
        join_tree: JoinTree,
        registry: Registry,
        dialect: str = "duckdb",
    ):
        self._select = select
        self._base_model = base_model
        self._registry = registry
        self._join_tree = join_tree
        self._already_joined: set[str] = set()
        self._already_joined.add(base_model.name)
        self._dialect = dialect

    def join_reference(self, reference: AttributePath):
        """Join the reference to the current base model"""
        traverser = reference.traverser()

        if self._base_model.name != reference.base_model:
            # Join to the base_model
            self._join(
                from_model_name=self._base_model.name,
                from_table_alias=traverser.alias(self._base_model.name),
                to_model_name=reference.base_model,
                create_alias=traverser.alias,
            )

        from_model_name = reference.base_model
        from_table_alias = traverser.alias(reference.base_model)
        from_table_through_attribute = traverser.current_attribute_name

        while traverser.next():
            self._join(
                from_model_name=from_model_name,
                from_table_alias=from_table_alias,
                to_model_name=traverser.current_model_name,
                create_alias=traverser.alias,
                through_attribute=from_table_through_attribute,
            )

            from_model_name = traverser.current_model_name
            from_table_alias = traverser.alias(from_model_name)
            from_table_through_attribute = traverser.current_attribute_name

    def _join(
        self,
        *,
        from_model_name: str,
        from_table_alias: str,
        to_model_name: str,
        create_alias: t.Callable[[str], str] = lambda x: x,
        through_attribute: str = "",
    ):
        logger.debug(
            f"Joining `{from_model_name}` to `{to_model_name}` through `{through_attribute}`"
        )
        registry = self._registry
        query = self._select

        join_path = self._join_relationships(
            from_model_name, to_model_name, through_attribute=through_attribute
        )

        logger.debug(f"Join path: {join_path}")

        for bound_relationship in join_path:
            referenced_model = registry.get_model(
                bound_relationship.relationship.ref_model
            )
            referenced_model_alias = create_alias(
                bound_relationship.relationship.ref_model
            )

            referenced_model_table = referenced_model.table_exp.as_(
                referenced_model_alias
            )

            if referenced_model_alias in self._already_joined:
                continue

            query = query.join(
                referenced_model_table,
                on=" AND ".join(
                    [
                        f"{from_table_alias}.{exp.to_identifier(source_foreign_key)} = {referenced_model_alias}.{exp.to_identifier(ref_key)}"
                        for source_foreign_key, ref_key in zip(
                            bound_relationship.relationship.source_foreign_key,
                            bound_relationship.relationship.ref_key,
                        )
                    ]
                ),
                join_type="left",
            )

            from_table_alias = referenced_model_alias
            from_model_name = referenced_model.name

            self._already_joined.add(referenced_model_alias)
        self._select = query

    def _join_relationships(
        self, from_model: str, to_model: str, through_attribute: str = ""
    ) -> t.List[BoundRelationship]:
        """Returns the join path between two models"""
        path = self._join_tree.get_path(from_model, to_model)

        def build_join_path(
            model_path: t.List[str], via_attribute: str = ""
        ) -> t.List[BoundRelationship]:
            prev_model: Model | None = None
            join_path: t.List[BoundRelationship] = []
            for model_name in model_path:
                if prev_model is None:
                    via_attribute = via_attribute
                    prev_model = self._registry.models[model_name]
                    continue

                relationship = prev_model.find_relationship(
                    name=via_attribute, model_ref=self._registry.models[model_name]
                )
                prev_model = self._registry.models[model_name]
                join_path.append(relationship)
            return join_path

        return build_join_path(path, through_attribute)

    @property
    def joined_query(self):
        return self._select



import os
import typing as t

import duckdb
import pandas as pd
import pytest

from .definition import Registry
from .testing import setup_registry

CURR_DIR = os.path.dirname(os.path.abspath(__file__))
FIXTURES_DIR = os.path.join(CURR_DIR, "fixtures")


@pytest.fixture
def semantic_registry() -> t.Iterator[Registry]:
    yield setup_registry()


@pytest.fixture
def semantic_db_conn() -> t.Iterator[duckdb.DuckDBPyConnection]:
    conn = duckdb.connect()

    csvs = [
        "artifacts_v1",
        "projects_v1",
        "collections_v1",
        "projects_by_collection_v1",
        "artifacts_by_project_v1",
        "artifacts_by_collection_v1",
        "int_events__github",
    ]

    conn.sql("CREATE SCHEMA oso;")

    for csv in csvs:
        path = os.path.join(FIXTURES_DIR, f"{csv}.csv")
        df = pd.read_csv(path)  # noqa: F841
        if csv == "int_events__github":
            print(df)
        table_name = f"oso.{csv}"
        conn.sql(f"CREATE OR REPLACE TABLE {table_name} AS SELECT * FROM df;")

    yield conn



from sqlglot import exp


def hash_expressions(
    *expressions: exp.Expression, dialect: str = "duckdb"
) -> exp.Expression:
    """Given the expressions return a sql expression that hashes them together.

    This is useful for creating a deterministic ID from multiple columns or values.
    """
    # Convert all expressions to strings
    expressions_cast_to_str = [
        exp.Cast(this=expr, to=exp.DataType.build("VARCHAR")) for expr in expressions
    ]
    concatenated = exp.Concat(
        expressions=expressions_cast_to_str, safe=True, coalesce=False
    )
    if dialect == "trino":
        # Trino's SHA256 function only accepts type `varbinary`. So we convert
        # the varchar to varbinary with trino's to_utf8.
        concatenated = exp.Anonymous(this="to_utf8", expressions=[concatenated])
    sha = exp.SHA2(
        this=concatenated,
        length=exp.Literal(this=256, is_string=False),
    )
    return sha


def exp_to_str(expression: exp.Expression | str) -> str:
    if isinstance(expression, str):
        return expression
    if isinstance(expression, exp.Literal):
        return str(expression.this)
    elif isinstance(expression, exp.Identifier):
        return str(expression.this)
    else:
        return str(expression)



# ruff: noqa: F401 F403

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pyoso")
except PackageNotFoundError:
    __version__ = "unknown"

from .definition import *
from .errors import InvalidAttributeReferenceError
from .query import QueryBuilder
from .register import register_oso_models



import textwrap

from ..definition import (
    Dimension,
    Measure,
    Model,
    Registry,
    Relationship,
    RelationshipType,
)


def register_entities(registry: Registry, catalog_name: str = "iceberg"):
    # ============================================"
    # CORE ENTITIES
    # ============================================"

    registry.register(
        Model(
            name="artifacts",
            table=f"{catalog_name}.oso.artifacts_v1",
            description=textwrap.dedent(
                """
                An artifact. This is the smallest atom of an acting entity in
                OSO. Artifacts are usually repositories, blockchain addresses,
                or some representation of a user. Artifacts do not generally
                represent a group of any kind, but rather a single entity.
                """
            ),
            dimensions=[
                Dimension(
                    name="artifact_id",
                    description="The unique identifier for the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="artifact_source_id",
                    description="The native identifier for the artifact from the source, such as a GitHub repository ID or NPM package ID",
                    column_name="artifact_source_id",
                ),
                Dimension(
                    name="artifact_source",
                    description='The original source of the artifact, such as "GITHUB", "NPM", or "DEFILLAMA"',
                    column_name="artifact_source",
                ),
                Dimension(
                    name="artifact_namespace",
                    description="The grouping or namespace of the artifact, such as the GitHub organization or NPM scope. It will will be empty if the artifact source does not have its own namespacing conventions",
                    column_name="artifact_namespace",
                ),
                Dimension(
                    name="artifact_name",
                    description="The name of the artifact, such as the GitHub repository name, NPM package name, or blockchain address",
                    column_name="artifact_name",
                ),
            ],
            primary_key="artifact_id",
            relationships=[
                Relationship(
                    name="by_project",
                    source_foreign_key="artifact_id",
                    ref_model="artifacts_by_project",
                    ref_key="artifact_id",
                    type=RelationshipType.ONE_TO_MANY,
                ),
                Relationship(
                    name="by_collection",
                    source_foreign_key="artifact_id",
                    ref_model="artifacts_by_collection",
                    ref_key="artifact_id",
                    type=RelationshipType.ONE_TO_MANY,
                ),
            ],
            measures=[
                Measure(
                    name="count",
                    description="The number of artifacts",
                    query="COUNT(self.artifact_id)",
                ),
                Measure(
                    name="distinct_count",
                    description="The number of distinct artifacts",
                    query="COUNT(DISTINCT self.artifact_id)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="projects",
            table=f"{catalog_name}.oso.projects_v1",
            description=textwrap.dedent(
                """
                A project is a collection of related artifacts. A project is
                usually, but not limited to, some kind of organization, company,
                or group that controls a set of artifacts.
                """
            ),
            dimensions=[
                Dimension(
                    name="project_id",
                    description="The unique identifier for the project, as generated by OSO",
                    column_name="project_id",
                ),
                Dimension(
                    name="project_source",
                    description='The source of the project, such as "OSS_DIRECTORY", or "OP_ATLAS"',
                    column_name="project_source",
                ),
                Dimension(
                    name="project_namespace",
                    description="The namespace of the project within the source, which may include organization information",
                    column_name="project_namespace",
                ),
                Dimension(
                    name="project_name",
                    description="the project slug or other name given to the project (unique to the registry but not globally unique)",
                    column_name="project_name",
                ),
                Dimension(
                    name="display_name",
                    description="The display name of the project, which may be a more human-readable name than the project_name",
                    column_name="display_name",
                ),
                Dimension(
                    name="description",
                    description="A short description of the project",
                    column_name="description",
                ),
            ],
            primary_key="project_id",
            relationships=[
                Relationship(
                    name="by_collection",
                    source_foreign_key="project_id",
                    ref_model="projects_by_collection",
                    ref_key="project_id",
                    type=RelationshipType.ONE_TO_MANY,
                ),
            ],
            measures=[
                Measure(
                    name="count",
                    description="The number of projects",
                    query="COUNT(self.project_id)",
                ),
                Measure(
                    name="distinct_count",
                    description="The number of distinct projects",
                    query="COUNT(DISTINCT self.project_id)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="collections",
            table=f"{catalog_name}.oso.collections_v1",
            description=textwrap.dedent(
                """
                A collection is a group of related projects. A collection is an
                arbitrary grouping of projects. Sometimes these groupings are
                used to group things together by some common dependency tree or
                some specific community known to OSO.
                """
            ),
            dimensions=[
                Dimension(
                    name="collection_id",
                    description="The unique identifier for the collection",
                    column_name="collection_id",
                ),
                Dimension(
                    name="collection_source",
                    description='The source of the collection, such as "OP_ATLAS" or "OSS_DIRECTORY"',
                    column_name="collection_source",
                ),
                Dimension(
                    name="collection_namespace",
                    description="The grouping or namespace of the collection",
                    column_name="collection_namespace",
                ),
                Dimension(
                    name="collection_name",
                    description="The name of the collection, such as an ecosystem name. If the collections are from OSS_DIRECTORY, this will be a unique collection slug",
                    column_name="collection_name",
                ),
                Dimension(
                    name="display_name",
                    description="The display name of the collection, which may not be unique. This is typically a human-readable name",
                    column_name="display_name",
                ),
                Dimension(
                    name="description",
                    description="brief summary or purpose of the collection",
                    column_name="description",
                ),
            ],
            primary_key="collection_id",
            measures=[
                Measure(
                    name="count",
                    description="The number of collections",
                    query="COUNT(self.collection_id)",
                ),
                Measure(
                    name="distinct_count",
                    description="The number of distinct collections",
                    query="COUNT(DISTINCT self.collection_id)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="contracts",
            table=f"{catalog_name}.oso.contracts_v0",
            description=textwrap.dedent(
                """
                This table contains information about smart contracts deployed across 
                various blockchain networks, including deployment details and relationships.
                """
            ),
            dimensions=[
                Dimension(
                    name="deployment_date",
                    description="The date of a contract deployment",
                    column_name="deployment_date",
                ),
                Dimension(
                    name="contract_address",
                    description="The address of the contract",
                    column_name="contract_address",
                ),
                Dimension(
                    name="contract_namespace",
                    description="The chain of the contract",
                    column_name="contract_namespace",
                ),
                Dimension(
                    name="originating_address",
                    description="The EOA address that initiated the contract deployment transaction",
                    column_name="originating_address",
                ),
                Dimension(
                    name="factory_address",
                    description="The address of the factory that deployed the contract, if this is the same as the originating address then this was a direct deployment by an EOA",
                    column_name="factory_address",
                ),
                Dimension(
                    name="root_deployer_address",
                    description="The EOA address that is considered the root deployer of the contract. If the contract was deployed directly by an EOA, this should be the same as the originating address, if the contract was deployed by a factory this is the creator of the factory",
                    column_name="root_deployer_address",
                ),
                Dimension(
                    name="sort_weight",
                    description="A weight used for sorting contracts. At this time, this is the tx count of the last 180 days",
                    column_name="sort_weight",
                ),
            ],
            primary_key="contract_address",
            measures=[
                Measure(
                    name="count",
                    description="The number of contracts",
                    query="COUNT(self.contract_address)",
                ),
                Measure(
                    name="distinct_count",
                    description="The number of distinct contracts",
                    query="COUNT(DISTINCT self.contract_address)",
                ),
                Measure(
                    name="avg_sort_weight",
                    description="The average sort weight of contracts",
                    query="AVG(self.sort_weight)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="repositories",
            table=f"{catalog_name}.oso.repositories_v0",
            description=textwrap.dedent(
                """
                This table contains information about source code repositories,
                primarily from GitHub, including repository metadata and statistics.
                """
            ),
            dimensions=[
                Dimension(
                    name="artifact_id",
                    description="The unique identifier for the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="artifact_source_id",
                    description="The native identifier for the artifact from the source",
                    column_name="artifact_source_id",
                ),
                Dimension(
                    name="artifact_source",
                    description="The original source of the artifact",
                    column_name="artifact_source",
                ),
                Dimension(
                    name="artifact_namespace",
                    description="The grouping or namespace of the artifact",
                    column_name="artifact_namespace",
                ),
                Dimension(
                    name="artifact_name",
                    description="The name of the artifact",
                    column_name="artifact_name",
                ),
                Dimension(
                    name="artifact_url",
                    description="The URL of the repository",
                    column_name="artifact_url",
                ),
                Dimension(
                    name="is_fork",
                    description="Whether the repository is a fork",
                    column_name="is_fork",
                ),
                Dimension(
                    name="branch",
                    description="The default branch of the repository",
                    column_name="branch",
                ),
                Dimension(
                    name="star_count",
                    description="The number of stars for the repository",
                    column_name="star_count",
                ),
                Dimension(
                    name="watcher_count",
                    description="The number of watchers for the repository",
                    column_name="watcher_count",
                ),
                Dimension(
                    name="fork_count",
                    description="The number of forks for the repository",
                    column_name="fork_count",
                ),
                Dimension(
                    name="license_name",
                    description="The name of the license",
                    column_name="license_name",
                ),
                Dimension(
                    name="license_spdx_id",
                    description="The SPDX ID of the license",
                    column_name="license_spdx_id",
                ),
                Dimension(
                    name="language",
                    description="The primary programming language of the repository",
                    column_name="language",
                ),
                Dimension(
                    name="created_at",
                    description="The timestamp when the repository was created",
                    column_name="created_at",
                ),
                Dimension(
                    name="updated_at",
                    description="The timestamp when the repository was last updated",
                    column_name="updated_at",
                ),
            ],
            primary_key="artifact_id",
            measures=[
                Measure(
                    name="count",
                    description="The number of repositories",
                    query="COUNT(*)",
                ),
                Measure(
                    name="total_stars",
                    description="The total number of stars across repositories",
                    query="SUM(self.star_count)",
                ),
                Measure(
                    name="avg_stars",
                    description="The average number of stars per repository",
                    query="AVG(self.star_count)",
                ),
                Measure(
                    name="total_forks",
                    description="The total number of forks across repositories",
                    query="SUM(self.fork_count)",
                ),
                Measure(
                    name="avg_forks",
                    description="The average number of forks per repository",
                    query="AVG(self.fork_count)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="package_owners",
            table=f"{catalog_name}.oso.package_owners_v0",
            description=textwrap.dedent(
                """
                This table contains information about package ownership relationships
                between packages and their owners across different package registries.
                """
            ),
            dimensions=[
                Dimension(
                    name="package_project_id",
                    description="The project ID of the package",
                    column_name="package_project_id",
                ),
                Dimension(
                    name="package_artifact_id",
                    description="The artifact ID of the package",
                    column_name="package_artifact_id",
                ),
                Dimension(
                    name="package_artifact_source",
                    description="The source of the package artifact",
                    column_name="package_artifact_source",
                ),
                Dimension(
                    name="package_artifact_namespace",
                    description="The namespace of the package artifact",
                    column_name="package_artifact_namespace",
                ),
                Dimension(
                    name="package_artifact_name",
                    description="The name of the package artifact",
                    column_name="package_artifact_name",
                ),
                Dimension(
                    name="package_owner_project_id",
                    description="The project ID of the package owner",
                    column_name="package_owner_project_id",
                ),
                Dimension(
                    name="package_owner_artifact_id",
                    description="The artifact ID of the package owner",
                    column_name="package_owner_artifact_id",
                ),
                Dimension(
                    name="package_owner_source",
                    description="The source of the package owner",
                    column_name="package_owner_source",
                ),
                Dimension(
                    name="package_owner_artifact_namespace",
                    description="The namespace of the package owner artifact",
                    column_name="package_owner_artifact_namespace",
                ),
                Dimension(
                    name="package_owner_artifact_name",
                    description="The name of the package owner artifact",
                    column_name="package_owner_artifact_name",
                ),
            ],
            primary_key=["package_artifact_id", "package_owner_artifact_id"],
            measures=[
                Measure(
                    name="count",
                    description="The number of package ownership records",
                    query="COUNT(*)",
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="sboms",
            table=f"{catalog_name}.oso.sboms_v0",
            description=textwrap.dedent(
                """
                Software Bill of Materials (SBOM) data containing information about
                software dependencies between projects and packages.
                """
            ),
            dimensions=[
                Dimension(
                    name="from_project_id",
                    description="The project ID of the dependent project",
                    column_name="from_project_id",
                ),
                Dimension(
                    name="from_artifact_id",
                    description="The artifact ID of the dependent artifact",
                    column_name="from_artifact_id",
                ),
                Dimension(
                    name="from_artifact_source",
                    description="The source of the dependent artifact",
                    column_name="from_artifact_source",
                ),
                Dimension(
                    name="from_artifact_namespace",
                    description="The namespace of the dependent artifact",
                    column_name="from_artifact_namespace",
                ),
                Dimension(
                    name="from_artifact_name",
                    description="The name of the dependent artifact",
                    column_name="from_artifact_name",
                ),
                Dimension(
                    name="to_package_project_id",
                    description="The project ID of the dependency package",
                    column_name="to_package_project_id",
                ),
                Dimension(
                    name="to_package_artifact_id",
                    description="The artifact ID of the dependency package",
                    column_name="to_package_artifact_id",
                ),
                Dimension(
                    name="to_package_artifact_source",
                    description="The source of the dependency package",
                    column_name="to_package_artifact_source",
                ),
                Dimension(
                    name="to_package_artifact_namespace",
                    description="The namespace of the dependency package",
                    column_name="to_package_artifact_namespace",
                ),
                Dimension(
                    name="to_package_artifact_name",
                    description="The name of the dependency package",
                    column_name="to_package_artifact_name",
                ),
            ],
            primary_key=["from_artifact_id", "to_package_artifact_id"],
            measures=[
                Measure(
                    name="count",
                    description="The number of SBOM dependency records",
                    query="COUNT(*)",
                ),
                Measure(
                    name="distinct_from_artifacts",
                    description="The number of distinct artifacts with dependencies",
                    query="COUNT(DISTINCT self.from_artifact_id)",
                ),
                Measure(
                    name="distinct_to_packages",
                    description="The number of distinct dependency packages",
                    query="COUNT(DISTINCT self.to_package_artifact_id)",
                ),
            ],
        )
    )

    # ============================================"
    # RELATIONSHIP TABLES
    # ============================================"

    registry.register(
        Model(
            name="artifacts_by_project",
            table=f"{catalog_name}.oso.artifacts_by_project_v1",
            description=textwrap.dedent(
                """
                The join table between artifacts and projects. This table
                represents the many-to-many relationship between artifacts and
                projects.
                """
            ),
            dimensions=[
                Dimension(
                    name="artifact_id",
                    description="The unique identifier for the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="artifact_source_id",
                    description="The native identifier for the artifact from the source, such as a GitHub repository ID or NPM package ID",
                    column_name="artifact_source_id",
                ),
                Dimension(
                    name="artifact_source",
                    description='The original source of the artifact, such as "GITHUB", "NPM", or "DEFILLAMA"',
                    column_name="artifact_source",
                ),
                Dimension(
                    name="artifact_namespace",
                    description="The grouping or namespace of the artifact, such as the GitHub organization or NPM scope",
                    column_name="artifact_namespace",
                ),
                Dimension(
                    name="artifact_name",
                    description="The name of the artifact, such as the GitHub repository name, NPM package name, or blockchain address",
                    column_name="artifact_name",
                ),
                Dimension(
                    name="project_id",
                    description="The unique identifier for the project that owns the artifact",
                    column_name="project_id",
                ),
                Dimension(
                    name="project_source",
                    description='The source of the project, such as "OP_ATLAS" or "OSS_DIRECTORY"',
                    column_name="project_source",
                ),
                Dimension(
                    name="project_namespace",
                    description="The grouping or namespace of the project",
                    column_name="project_namespace",
                ),
                Dimension(
                    name="project_name",
                    description="The name of the project, such as the GitHub organization name. If the projects are from OSS_DIRECTORY, this will be a unique project slug",
                    column_name="project_name",
                ),
            ],
            primary_key=["artifact_id", "project_id"],
            relationships=[
                Relationship(
                    source_foreign_key="project_id",
                    ref_model="projects",
                    ref_key="project_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="projects_by_collection",
            table=f"{catalog_name}.oso.projects_by_collection_v1",
            description=textwrap.dedent(
                """
                The join table between projects and collections. This table
                represents the many-to-many relationship between projects and
                collections.
                """
            ),
            dimensions=[
                Dimension(
                    name="project_id",
                    description="The unique identifier for the project, as generated by OSO",
                    column_name="project_id",
                ),
                Dimension(
                    name="project_source",
                    description='The source of the project, such as "OSS_DIRECTORY", or "OP_ATLAS"',
                    column_name="project_source",
                ),
                Dimension(
                    name="project_namespace",
                    description="The namespace of the project within the source, which may include organization information",
                    column_name="project_namespace",
                ),
                Dimension(
                    name="project_name",
                    description="The project slug or other name given to the project (unique to the registry but not globally unique)",
                    column_name="project_name",
                ),
                Dimension(
                    name="collection_id",
                    description="The unique identifier for the collection, as generated by OSO",
                    column_name="collection_id",
                ),
                Dimension(
                    name="collection_source",
                    description='The source of the collection, such as "OSS_DIRECTORY", or "OP_ATLAS"',
                    column_name="collection_source",
                ),
                Dimension(
                    name="collection_namespace",
                    description="The namespace of the collection within the source, which may include organization information",
                    column_name="collection_namespace",
                ),
                Dimension(
                    name="collection_name",
                    description="The name of the collection",
                    column_name="collection_name",
                ),
            ],
            primary_key=["project_id", "collection_id"],
            relationships=[
                Relationship(
                    source_foreign_key="collection_id",
                    ref_model="collections",
                    ref_key="collection_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="project_id",
                    ref_model="projects",
                    ref_key="project_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )

    registry.register(
        Model(
            name="artifacts_by_collection",
            table=f"{catalog_name}.oso.artifacts_by_collection_v1",
            description=textwrap.dedent(
                """
                The join table between artifacts and collections. This table represents 
                the many-to-many relationship between artifacts and collections.
                """
            ),
            dimensions=[
                Dimension(
                    name="artifact_id",
                    description="The unique identifier for the artifact",
                    column_name="artifact_id",
                ),
                Dimension(
                    name="artifact_source_id",
                    description="The native identifier for the artifact from the source, such as a GitHub repository ID or NPM package ID",
                    column_name="artifact_source_id",
                ),
                Dimension(
                    name="artifact_source",
                    description='The original source of the artifact, such as "GITHUB", "NPM", or "DEFILLAMA"',
                    column_name="artifact_source",
                ),
                Dimension(
                    name="artifact_namespace",
                    description="The grouping or namespace of the artifact, such as the GitHub organization or NPM scope",
                    column_name="artifact_namespace",
                ),
                Dimension(
                    name="artifact_name",
                    description="The name of the artifact, such as the GitHub repository name, NPM package name, or blockchain address",
                    column_name="artifact_name",
                ),
                Dimension(
                    name="collection_id",
                    description="The unique identifier for the collection that includes the artifact",
                    column_name="collection_id",
                ),
                Dimension(
                    name="collection_source",
                    description='The source of the collection, such as "OP_ATLAS" or "OSS_DIRECTORY"',
                    column_name="collection_source",
                ),
                Dimension(
                    name="collection_namespace",
                    description="The grouping or namespace of the collection",
                    column_name="collection_namespace",
                ),
                Dimension(
                    name="collection_name",
                    description="The name of the collection, such as an ecosystem name. If the collections are from OSS_DIRECTORY, this will be a unique collection slug",
                    column_name="collection_name",
                ),
            ],
            primary_key=["artifact_id", "collection_id"],
            relationships=[
                Relationship(
                    source_foreign_key="collection_id",
                    ref_model="collections",
                    ref_key="collection_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
                Relationship(
                    source_foreign_key="artifact_id",
                    ref_model="artifacts",
                    ref_key="artifact_id",
                    type=RelationshipType.MANY_TO_ONE,
                ),
            ],
        )
    )



import duckdb
import pandas as pd
from sqlglot import parse_one

from .definition import (
    AttributePath,
    BoundMeasure,
    Dimension,
    Measure,
    Model,
    ModelHasNoJoinPath,
    Registry,
    Relationship,
    RelationshipType,
    SemanticExpression,
)
from .testing import setup_registry


def test_attribute_reference_traversal():
    ref1 = AttributePath.from_string("a.b->c.d->e.f")
    ref2 = AttributePath(path=["a.b", "c.d", "e.f"])
    ref3 = AttributePath(path=["a.b", "c.d", "x.y"])
    ref4 = AttributePath(path=["a.g", "c.d", "e.f"])
    assert ref1 == ref2
    assert ref1 != ref3

    t1 = ref1.traverser()
    t2 = ref2.traverser()

    assert t1.current_model_name == t2.current_model_name
    assert t1.current_table_alias == t2.current_table_alias

    while t1.next() and t2.next():
        assert t1.current_model_name == t2.current_model_name
        assert t1.current_table_alias == t2.current_table_alias
        assert t1.current_attribute_name == t2.current_attribute_name
        assert t1.alias("foo") == t2.alias("foo")

    while t1.prev():
        pass

    t3 = ref3.traverser()

    t1.next()
    t3.next()

    assert t1.current_table_alias == t3.current_table_alias
    print(t1.alias_stack)
    print(t3.alias_stack)
    assert t1.alias("foo") == t3.alias("foo")

    t1.next()
    t3.next()

    assert t1.current_attribute_name != t3.current_attribute_name
    assert t1.alias("foo") == t3.alias("foo")

    while t1.prev():
        pass

    t4 = ref4.traverser()

    assert t1.alias("foo") == t4.alias("foo")

    t1.next()
    t4.next()

    assert t1.alias("foo") != t4.alias("foo")


def test_semantic_model_shortest_path():
    registry = setup_registry()

    artifact_project_tree = registry.dag.find_best_join_tree(
        [
            AttributePath(
                path=[
                    "artifacts.artifact_id",
                    "artifacts_by_project.project_id",
                    "projects.artifact_id",
                ]
            ),
        ]
    )
    github_artifact_tree = registry.dag.find_best_join_tree(
        [
            AttributePath.from_string("int_events__github.to->artifacts.artifact_id"),
        ]
    )
    github_collection_tree = registry.dag.find_best_join_tree(
        [
            AttributePath.from_string(
                "int_events__github.to->collections.collection_id"
            ),
        ]
    )

    print(github_collection_tree.root)
    print(github_collection_tree.parents)

    assert github_artifact_tree.get_path("int_events__github", "artifacts") == [
        "int_events__github",
        "artifacts",
    ]

    assert github_artifact_tree.get_path("artifacts", "int_events__github") == [
        "artifacts",
        "int_events__github",
    ]

    assert artifact_project_tree.get_path("artifacts", "projects") == [
        "artifacts",
        "artifacts_by_project",
        "projects",
    ]

    assert github_collection_tree.get_path("int_events__github", "collections") == [
        "int_events__github",
        "artifacts",
        "artifacts_by_collection",
        "collections",
    ]

    to_artifact_ref = registry.get_model("int_events__github").find_relationship(
        model_ref=registry.get_model("artifacts"), name="to"
    )
    from_artifact_ref = registry.get_model("int_events__github").find_relationship(
        model_ref=registry.get_model("artifacts"), name="from"
    )

    assert to_artifact_ref.relationship.ref_model == "artifacts"
    assert to_artifact_ref.relationship.name == "to"
    assert to_artifact_ref.relationship.source_foreign_key[0] == "to_artifact_id"
    assert from_artifact_ref.relationship.ref_model == "artifacts"
    assert from_artifact_ref.relationship.name == "from"
    assert from_artifact_ref.relationship.source_foreign_key[0] == "from_artifact_id"


def test_semantic_query(semantic_db_conn: duckdb.DuckDBPyConnection):
    registry = setup_registry()

    query = registry.select(
        "int_events__github.event_type AS event_type",
        "int_events__github.to->collections.collection_name AS collection_name",
        "int_events__github.total_amount AS total_amount",
    ).where("int_events__github.from->artifacts.artifact_name = 'repo1'")

    query_exp = query.build()
    assert query_exp is not None
    print(query_exp.sql(dialect="duckdb", pretty=True))

    result = semantic_db_conn.sql(query_exp.sql(dialect="duckdb"))
    result_df = result.df()
    result_df = result_df.sort_values(by=["collection_name"])
    result_df = result_df.reset_index(drop=True)
    print(result_df)

    pd.testing.assert_frame_equal(
        result_df,
        pd.DataFrame(
            {
                "event_type": ["COMMIT_CODE"],
                "collection_name": ["collection1"],
                "total_amount": [7.0],
            }
        ),
        check_like=True,
    )


def test_resolve_attributes(semantic_registry: Registry):
    # Test resolving a single attribute
    ref0 = SemanticExpression(query="artifacts.artifact_name")

    resolved_expr0 = ref0.resolve(semantic_registry)
    assert resolved_expr0 is not None
    assert resolved_expr0 == parse_one("artifacts_e4e2753c.artifact_name"), (
        "Resolved expression should match expected SQL expression"
    )

    ref1 = SemanticExpression(query="int_events__github.to->artifacts.artifact_id")
    resolved_expr1 = ref1.resolve(semantic_registry)

    assert resolved_expr1 is not None
    assert resolved_expr1 == parse_one("artifacts_b52061f6.artifact_id")

    # Test resolving a more complex expression
    ref2 = SemanticExpression(query="int_events__github.month")
    resolved_expr2 = ref2.resolve(semantic_registry)
    assert resolved_expr2 is not None
    assert resolved_expr2 == parse_one(
        "DATE_TRUNC('month', int_events__github_022877ad.time)"
    )


def test_expand_reference():
    registry = setup_registry()

    ref = AttributePath.from_string("int_events__github.to->artifacts.artifact_id")
    expanded_refs = registry.expand_reference(ref)

    assert len(expanded_refs) == 1
    assert expanded_refs[0] == AttributePath.from_string(
        "int_events__github.to->artifacts.artifact_id"
    )

    # # Test with a more complex reference
    complex_ref = AttributePath.from_string("int_events__github.month")
    complex_expanded_refs = registry.expand_reference(complex_ref)
    print(complex_expanded_refs)
    assert len(complex_expanded_refs) == 2
    assert complex_expanded_refs == [
        AttributePath.from_string("int_events__github.month"),
        AttributePath.from_string("int_events__github.time"),
    ]


def test_attribute_reference_transformer():
    simple_column = SemanticExpression(query="github_event.to")

    references0 = simple_column.references()
    assert len(references0) == 1
    assert references0[0] == AttributePath.from_string("github_event.to")

    multiple_refs_sql = SemanticExpression(
        query="github_event.to->artifact.count > github_event.from->artifact.count"
    )

    references1 = multiple_refs_sql.references()

    assert len(references1) == 2

    assert AttributePath.from_string("github_event.to->artifact.count") in references1
    assert AttributePath.from_string("github_event.from->artifact.count") in references1


def test_resolve_measures():
    registry = Registry()

    model = Model(
        name="artifacts",
        table="oso.artifacts_v1",
        description="An artifact",
        dimensions=[
            Dimension(name="artifact_id", column_name="artifact_id"),
        ],
        primary_key="artifact_id",
        measures=[
            Measure(
                name="count",
                description="The number of artifacts",
                query="COUNT(self.artifact_id)",
            )
        ],
    )
    registry.register(model)

    measure = model.get_attribute("count")

    assert isinstance(measure, BoundMeasure), "count should be a BoundMetric"
    assert measure is not None, "metric should not be None"

    ref = AttributePath.from_string("artifacts.artifact_id")
    assert measure.references() == [ref], "Measure should reference artifact.id"


def test_dag_find_best_join_tree():
    """Test that find_best_join_tree returns the smallest possible tree."""
    from .definition import (
        AttributePath,
        Dimension,
        Model,
        Registry,
        Relationship,
        RelationshipType,
    )

    registry = Registry()

    # Create a simpler model graph for clearer testing:
    # A -> B -> D
    # A -> C -> D
    # This gives us two possible paths from A to D

    model_a = Model(
        name="model_a",
        table="table_a",
        relationships=[
            Relationship(
                name="to_b",
                ref_model="model_b",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="b_id",
                ref_key="id",
            ),
            Relationship(
                name="to_c",
                ref_model="model_c",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="c_id",
                ref_key="id",
            ),
        ],
        dimensions=[
            Dimension(name="id", description="A's ID"),
            Dimension(name="name", description="A's name"),
        ],
        measures=[],
    )

    model_b = Model(
        name="model_b",
        table="table_b",
        relationships=[
            Relationship(
                name="to_d",
                ref_model="model_d",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="d_id",
                ref_key="id",
            )
        ],
        dimensions=[
            Dimension(name="id", description="B's ID"),
            Dimension(name="name", description="B's name"),
        ],
        measures=[],
    )

    model_c = Model(
        name="model_c",
        table="table_c",
        relationships=[
            Relationship(
                name="to_d",
                ref_model="model_d",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="d_id",
                ref_key="id",
            )
        ],
        dimensions=[
            Dimension(name="id", description="C's ID"),
            Dimension(name="name", description="C's name"),
        ],
        measures=[],
    )

    model_d = Model(
        name="model_d",
        table="table_d",
        relationships=[],
        dimensions=[
            Dimension(name="id", description="D's ID"),
            Dimension(name="value", description="D's value"),
        ],
        measures=[],
    )

    # Register all models
    registry.register(model_a)
    registry.register(model_b)
    registry.register(model_c)
    registry.register(model_d)

    # Test 1: Simple case - only need A
    references_a_only = [AttributePath.from_string("model_a.name")]
    join_tree = registry.dag.find_best_join_tree(references_a_only)
    assert len(join_tree) == 1, f"Expected tree with 1 node, got {len(join_tree)}"
    assert join_tree.root == "model_a", (
        f"Expected root to be model_a, got {join_tree.root}"
    )

    # Test 2: Only need D
    references_d_only = [AttributePath.from_string("model_d.value")]
    join_tree = registry.dag.find_best_join_tree(references_d_only)
    assert len(join_tree) == 1, (
        f"Expected tree with 1 node (just D), got {len(join_tree)}"
    )
    assert join_tree.root == "model_d", (
        f"Expected root to be model_d, got {join_tree.root}"
    )

    # Test 3: Two models with direct relationship - A and B
    references_a_b = [
        AttributePath.from_string("model_a.name"),
        AttributePath.from_string("model_b.name"),
    ]
    join_tree = registry.dag.find_best_join_tree(references_a_b)
    assert len(join_tree) == 2, f"Expected tree with 2 nodes, got {len(join_tree)}"
    assert join_tree.root in [
        "model_a",
        "model_b",
    ], f"Expected root to be model_a or model_b, got {join_tree.root}"

    # Test 4: Path through A->...->D
    references_a_b_d = [
        AttributePath.from_string("model_a.name"),
        AttributePath.from_string("model_d.value"),
    ]
    join_tree = registry.dag.find_best_join_tree(references_a_b_d)
    expected_models = {"model_a", "model_d"}
    actual_models = set(join_tree.parents.keys())
    assert expected_models.issubset(actual_models), (
        f"Expected models {expected_models} to be subset of {actual_models}"
    )

    # Test 5: Verify that find_best_join_tree finds the minimal tree when multiple options exist
    references_both_paths = [
        AttributePath.from_string("model_a.name"),
        AttributePath.from_string("model_b.name"),
        AttributePath.from_string("model_d.value"),
    ]
    expected_models = {"model_a", "model_b", "model_d"}
    join_tree = registry.dag.find_best_join_tree(references_both_paths)
    actual_models = set(join_tree.parents.keys())
    assert expected_models.issubset(actual_models), (
        f"Expected models {expected_models} to be subset of {actual_models}"
    )


def test_dag_find_best_join_tree_comparison():
    """Test that find_best_join_tree chooses the smaller of two possible trees."""

    registry = Registry()

    # Create a simple linear chain: A -> B -> C
    model_a = Model(
        name="model_a",
        table="table_a",
        relationships=[
            Relationship(
                name="to_b",
                ref_model="model_b",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="b_id",
                ref_key="id",
            )
        ],
        dimensions=[
            Dimension(name="id", description="A's ID"),
            Dimension(name="name", description="A's name"),
        ],
        measures=[],
    )

    model_b = Model(
        name="model_b",
        table="table_b",
        relationships=[
            Relationship(
                name="to_c",
                ref_model="model_c",
                type=RelationshipType.MANY_TO_ONE,
                source_foreign_key="c_id",
                ref_key="id",
            )
        ],
        dimensions=[
            Dimension(name="id", description="B's ID"),
            Dimension(name="name", description="B's name"),
        ],
        measures=[],
    )

    model_c = Model(
        name="model_c",
        table="table_c",
        relationships=[],
        dimensions=[
            Dimension(name="id", description="C's ID"),
            Dimension(name="value", description="C's value"),
        ],
        measures=[],
    )

    # Register all models
    registry.register(model_a)
    registry.register(model_b)
    registry.register(model_c)

    # Test that when we only need B and C, we get a tree rooted at B or C (not A)
    references_b_c = [
        AttributePath.from_string("model_b.name"),
        AttributePath.from_string("model_b.to_c->model_c.value"),
    ]

    join_tree = registry.dag.find_best_join_tree(references_b_c)

    # Should only include B and C, not A
    assert len(join_tree) == 2, (
        f"Expected tree with 2 nodes (B,C), got {len(join_tree)}"
    )
    assert "model_a" not in join_tree.parents, (
        f"Tree should not include model_a: {join_tree.parents}"
    )
    assert "model_b" in join_tree.parents, (
        f"Tree should include model_b: {join_tree.parents}"
    )
    assert "model_c" in join_tree.parents, (
        f"Tree should include model_c: {join_tree.parents}"
    )

    print("✓ find_best_join_tree correctly excludes unnecessary models")


def test_dag_find_best_join_tree_no_path():
    """Test that find_best_join_tree raises appropriate error when no path exists."""
    registry = Registry()

    # Create two isolated models with no connections
    model_a = Model(
        name="model_a",
        table="table_a",
        relationships=[],
        dimensions=[
            Dimension(name="id", description="A's ID"),
            Dimension(name="name", description="A's name"),
        ],
        measures=[],
    )

    model_b = Model(
        name="model_b",
        table="table_b",
        relationships=[],
        dimensions=[
            Dimension(name="id", description="B's ID"),
            Dimension(name="name", description="B's name"),
        ],
        measures=[],
    )

    registry.register(model_a)
    registry.register(model_b)

    # Try to find a join tree that requires both isolated models
    references = [
        AttributePath.from_string("model_a.name"),
        AttributePath.from_string("model_b.name"),
    ]

    # This should raise ModelHasNoJoinPath exception
    try:
        registry.dag.find_best_join_tree(references)
        assert False, "Expected ModelHasNoJoinPath exception"
    except ModelHasNoJoinPath as e:
        assert "No join tree found" in str(e)
        print(
            "✓ find_best_join_tree correctly raises exception for disconnected models"
        )


def test_semantic_expression():
    registry = setup_registry()
    project_count_expression = SemanticExpression(query="projects.count > 10")

    assert project_count_expression.references() == [
        AttributePath.from_string("projects.count")
    ]

    final_expression = project_count_expression.resolve(registry)

    assert final_expression == parse_one("COUNT(projects_35a2864c.project_id) > 10")


def test_semantic_expression_with_cte(semantic_db_conn: duckdb.DuckDBPyConnection):
    registry = setup_registry()

    # Create a CTE for projects
    cte_query = registry.select("projects.project_id AS project_id").where(
        "projects.project_id = 'project1'"
    )
    cte_name = "project_1"

    query = registry.cte(cte_name, cte_query).select(
        "project_1.project_id AS project_id",
        "collections.collection_name AS collection_name",
    )
    query_exp = query.build()
    assert query_exp is not None
    print(query_exp.sql(dialect="duckdb", pretty=True))

    result = semantic_db_conn.sql(query_exp.sql(dialect="duckdb"))
    result_df = result.df()
    result_df = result_df.sort_values(by=["collection_name"])
    result_df = result_df.reset_index(drop=True)

    pd.testing.assert_frame_equal(
        result_df,
        pd.DataFrame(
            {
                "project_id": ["project1", "project1", "project1"],
                "collection_name": ["collection1", "collection2", "collection3"],
            }
        ),
        check_like=True,
    )


def test_semantic_expression_with_dependant_cte(
    semantic_db_conn: duckdb.DuckDBPyConnection,
):
    registry = setup_registry()

    # Create a CTE for projects
    cte_artifact_events = registry.select(
        "int_events__github.event_type AS event_type",
        "int_events__github.from->artifacts.artifact_name AS artifact_name",
        "int_events__github.from->artifacts.artifact_id AS artifact_id",
    )
    cte_project_events = registry.cte("artifact_events", cte_artifact_events).select(
        "artifact_events.event_type AS event_type",
        "artifact_events.artifact_name AS artifact_name",
        "projects.project_name AS project_name",
        "projects.project_id AS project_id",
    )
    query = (
        registry.cte("cte_project_events", cte_project_events)
        .select(
            "cte_project_events.project_id AS project_id",
            "cte_project_events.project_name AS project_name",
            "cte_project_events.event_type AS event_type",
            "collections.collection_name AS collection_name",
        )
        .where("collections.collection_name = 'collection1'")
    )
    query_exp = query.build()
    assert query_exp is not None
    print(query_exp.sql(dialect="duckdb", pretty=True))

    result = semantic_db_conn.sql(query_exp.sql(dialect="duckdb"))
    result_df = result.df()
    result_df = result_df.sort_values(by=["project_id"])
    result_df = result_df.reset_index(drop=True)

    pd.testing.assert_frame_equal(
        result_df,
        pd.DataFrame(
            {
                "project_id": ["project1", "project2"],
                "project_name": ["org1", "org2"],
                "event_type": ["COMMIT_CODE", "COMMIT_CODE"],
                "collection_name": ["collection1", "collection1"],
            }
        ),
        check_like=True,
    )
